
# slixmpp.xmlstream.xmlstream
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# This module provides the module for creating and
# interacting with generic XML streams, along with
# the necessary eventing infrastructure.
# Part of Slixmpp: The Slick XMPP Library
# :copyright: (c) 2011 Nathanael C. Fritz
# :license: MIT, see LICENSE for more details
from typing import (
    Any,
    Awaitable,
    Generator,
    Coroutine,
    Callable,
    Iterable,
    TypeVar,
    cast,
)

import asyncio
import functools
import logging
import socket as Socket
import ssl
import uuid
import warnings
import weakref
import collections

from contextlib import contextmanager
import xml.etree.ElementTree as ET
from asyncio import (
    AbstractEventLoop,
    BaseTransport,
    Future,
    Task,
    TimerHandle,
    Transport,
    wait,
)
from inspect import iscoroutinefunction
from pathlib import Path

from slixmpp.types import FilterString
from slixmpp.xmlstream.tostring import tostring
from slixmpp.xmlstream.stanzabase import StanzaBase, ElementBase
from slixmpp.xmlstream.resolver import resolve, default_resolver
from slixmpp.xmlstream.handler.base import BaseHandler

T = TypeVar('T')

#: The time in seconds to wait before timing out waiting for response stanzas.
RESPONSE_TIMEOUT = 30

log = logging.getLogger(__name__)


class ContinueQueue(Exception):
    """
    Exception raised in the send queue to "continue" from within an inner loop
    """


class NotConnectedError(Exception):
    """
    Raised when we try to send something over the wire but we are not
    connected.
    """


class InvalidCABundle(Exception):
    """
        Exception raised when the CA Bundle file hasn't been found.
    """

    def __init__(self, path: Path | Iterable[Path] | None):
        self.path = path


_T = TypeVar('_T', str, ElementBase, StanzaBase)


SyncFilter = Callable[[StanzaBase], StanzaBase | None]
AsyncFilter = Callable[[StanzaBase], Awaitable[StanzaBase | None]]


Filter = SyncFilter | AsyncFilter

_FiltersDict = dict[str, list[Filter]]

Handler = Callable[[Any], Any | Coroutine[Any, Any, Any]]


class XMLStream(asyncio.BaseProtocol):
    """
    An XML stream connection manager and event dispatcher.

    The XMLStream class abstracts away the issues of establishing a
    connection with a server and sending and receiving XML "stanzas".
    A stanza is a complete XML element that is a direct child of a root
    document element. Two streams are used, one for each communication
    direction, over the same socket. Once the connection is closed, both
    streams should be complete and valid XML documents.

    Three types of events are provided to manage the stream:
        :Stream: Triggered based on received stanzas, similar in concept
                 to events in a SAX XML parser.
        :Custom: Triggered manually.
        :Scheduled: Triggered based on time delays.

    Typically, stanzas are first processed by a stream event handler which
    will then trigger custom events to continue further processing,
    especially since custom event handlers may run in individual threads.

    :param socket: Use an existing socket for the stream. Defaults to
                   ``None`` to generate a new socket.
    :param string host: The name of the target server.
    :param int port: The port to use for the connection. Defaults to 0.
    """

    transport: Transport | None

    # The socket that is used internally by the transport object
    socket: ssl.SSLSocket | None

    # The backoff of the connect routine (increases exponentially
    # after each failure)
    _connect_loop_wait: float

    parser: ET.XMLPullParser | None
    xml_depth: int
    xml_root: ET.Element | None

    waiting_queue: asyncio.Queue

    # A dict of {name: handle}
    scheduled_events: dict[str, TimerHandle]

    ssl_context: ssl.SSLContext

    # The event to trigger when the create_connection() succeeds. It can
    # be "connected" or "tls_success" depending on the step we are at.
    event_when_connected: str

    #: The list of accepted ciphers, in OpenSSL Format.
    #: It might be useful to override it for improved security
    #: over the python defaults.
    ciphers: str | None

    #: Path to a file containing certificates for verifying the
    #: server SSL certificate. A non-``None`` value will trigger
    #: certificate checking.
    #:
    #: .. note::
    #:
    #:     On Mac OS X, certificates in the system keyring will
    #:     be consulted, even if they are not in the provided file.
    ca_certs: Path | Iterable[Path] | None

    #: Path to a file containing a client certificate to use for
    #: authenticating via SASL EXTERNAL. If set, there must also
    #: be a corresponding `:attr:keyfile` value.
    certfile: str | None

    #: Path to a file containing the private key for the selected
    #: client certificate to use for authenticating via SASL EXTERNAL.
    keyfile: str | None

    # The asyncio event loop
    _loop: AbstractEventLoop | None

    #: The default port to return when querying DNS records.
    default_port: int

    #: The domain to try when querying DNS records.
    default_domain: str

    #: The expected name of the server, for validation.
    _expected_server_name: str
    _service_name: str

    #: A custom address to connect to that was provided to connect(). When
    #: using DNS lookups, this is not used. Once set, it will be re-used until
    #: connect() is called with no parameters (or both host and port set to
    #: None).
    custom_address: tuple[str, int] | None

    #: Enable connecting to the server directly over SSL, in
    #: particular when the service provides two ports: one for
    #: non-SSL traffic and another for SSL traffic.
    enable_direct_tls: bool
    #: Enable connecting to the server using STARTTLS (i.e. upgrading a clear
    #: connection once established).
    enable_starttls: bool
    #: Enable connecting to the server using plaintext
    #: strongly not recommended (unless for localhost components)
    enable_plaintext: bool

    #: If set to ``True``, attempt to use IPv6.
    use_ipv6: bool

    #: If set to ``True``, allow using the ``dnspython`` DNS library
    #: if available. If set to ``False``, the builtin DNS resolver
    #: will be used, even if ``dnspython`` is installed.
    use_aiodns: bool

    #: Use CDATA for escaping instead of XML entities. Defaults
    #: to ``False``.
    use_cdata: bool

    #: The default namespace of the stream content, not of the
    #: stream wrapper it
    default_ns: str

    default_lang: str | None
    peer_default_lang: str | None

    #: The namespace of the enveloping stream element.
    stream_ns: str

    #: The default opening tag for the stream element.
    stream_header: str

    #: The default closing tag for the stream element.
    stream_footer: str

    #: If ``True``, periodically send a whitespace character over the
    #: wire to keep the connection alive. Mainly useful for connections
    #: traversing NAT.
    whitespace_keepalive: bool

    #: The default interval between keepalive signals when
    #: :attr:`whitespace_keepalive` is enabled.
    whitespace_keepalive_interval: int

    #: Flag for controlling if the session can be considered ended
    #: if the connection is terminated.
    end_session_on_disconnect: bool

    #: A mapping of XML namespaces to well-known prefixes.
    namespace_map: dict

    __root_stanza: list[type[StanzaBase]]
    __handlers: list[BaseHandler]
    __event_handlers: dict[str, list[tuple[Handler, bool]]]
    __filters: _FiltersDict

    # Current connection attempt (Future)
    _current_connection_attempt: Future | None

    #: The reason why we are disconnecting from the server
    disconnect_reason: str | None

    #: An asyncio Future being done when the stream is disconnected.
    disconnected: Future

    # If the session has been started or not
    _session_started: bool
    # If we want to bypass the send() check (e.g. unit tests)
    _always_send_everything: bool

    _run_out_filters: Future | None
    __slow_tasks: list[Task]
    __queued_stanzas: list[tuple[StanzaBase | str, bool]]

    #: List of DNS SRV services records which map to TLS services
    tls_services: set[str]
    #: List of DNS SRV services records which map to STARTTLS services
    starttls_services: set[str]

    def __init__(self, host: str = '', port: int = 0,
                 ssl_context: ssl.SSLContext | None = None,
                 loop: asyncio.AbstractEventLoop | None = None):
        self.transport = None
        self.socket = None
        self._connect_loop_wait = 0

        self.parser = None
        self.xml_depth = 0
        self.xml_root = None

        self.waiting_queue = asyncio.Queue()

        # A dict of {name: handle}
        self.scheduled_events = {}

        if ssl_context is None:
            self.ssl_context = ssl.create_default_context()
            self.ssl_context.check_hostname = True
            self.ssl_context.verify_mode = ssl.CERT_REQUIRED
            self.ssl_context.set_default_verify_paths()
        else:
            self.ssl_context = ssl_context

        self.event_when_connected = "connected"

        self.ciphers = None

        self.ca_certs = None

        self.keyfile = None

        if loop:
            self._loop = loop
        else:
            self._loop = None
            # Let the loop be set by the getter
            self.loop


        self.default_port = int(port)
        self.default_domain = ''

        self._expected_server_name = ''
        self._service_name = ''

        self.custom_address = None

        self.enable_starttls = True
        self.enable_direct_tls = True
        self.enable_plaintext = False

        self.tls_services = set()
        self.starttls_services = set()
        self.use_ipv6 = True

        self.use_aiodns = True
        self.use_cdata = False

        self.default_ns = ''

        self.default_lang = None
        self.peer_default_lang = None

        self.stream_ns = ''
        self.stream_header = "<stream>"
        self.stream_footer = "</stream>"

        self.whitespace_keepalive = True
        self.whitespace_keepalive_interval = 300

        self.end_session_on_disconnect = True
        self.namespace_map = {StanzaBase.xml_ns: 'xml'}

        self.__root_stanza = []
        self.__handlers = []
        self.__event_handlers = {}
        self.__filters = {
            'in': [], 'out': [], 'out_sync': []
        }

        self._current_connection_attempt = None

        self.disconnect_reason = None
        self.disconnected = Future(loop=self.loop)
        self._session_started = False
        self._always_send_everything = False

        self.add_event_handler('disconnected', self._remove_schedules)
        self.add_event_handler('disconnected', self._set_disconnected)
        self.add_event_handler('session_start', self._start_keepalive)
        self.add_event_handler('session_start', self._set_session_start)
        self.add_event_handler('session_resumed', self._set_session_start)

        self._run_out_filters = None
        self.__slow_tasks = []
        self.__queued_stanzas = []

    @property
    def loop(self) -> AbstractEventLoop:
        if self._loop is None:
            try:
                # Python < 3.14
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    self._loop = asyncio.get_event_loop()
            except RuntimeError:
                try:
                    current = asyncio.get_running_loop()
                except RuntimeError:
                    current = None
                if current is not None:
                    self._loop = current
                else:
                    loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(loop)
                    self._loop = loop
        return self._loop

    @loop.setter
    def loop(self, value: AbstractEventLoop) -> None:
        self._loop = value

    def new_id(self) -> str:
        """Generate and return a new stream ID in hexadecimal form.

        Many stanzas, handlers, or matchers may require unique
        ID values. Using this method ensures that all new ID values
        are unique in this stream.
        """
        return uuid.uuid4().hex

    def _set_session_start(self, event: Any) -> None:
        """
        On session start, queue all pending stanzas to be sent.
        """
        self._session_started = True
        for stanza in self.__queued_stanzas:
            self.waiting_queue.put_nowait(stanza)
        self.__queued_stanzas = []

    def _set_disconnected(self, event: Any) -> None:
        self._session_started = False

    def _set_disconnected_future(self) -> None:
        """Set the self.disconnected future on disconnect"""
        if not self.disconnected.done():
            self.disconnected.set_result(True)
        self.disconnected = asyncio.Future(loop=self.loop)

    def connect(self, host: str | None = None, port: int | None = None) -> asyncio.Future:
        """Create a new socket and connect to the server.

        :param host: The name of the desired server for the connection.
        :param port: Port to connect to on the server.
        :returns: A future on the current connection attempt
        """
        if self._run_out_filters is None or self._run_out_filters.done():
            self._run_out_filters = asyncio.ensure_future(
                self.run_filters(),
                loop=self.loop,
            )

        self.disconnect_reason = None
        self.cancel_connection_attempt()
        self._connect_loop_wait = 0
        if host and port:
            self.custom_address = (host, int(port))
        elif (host, port) == (None, None):
            self.custom_address = None

        self.event("connecting")
        self._current_connection_attempt = asyncio.ensure_future(
            self._connect_loop(),
            loop=self.loop,
        )
        return self._current_connection_attempt

    async def _connect_loop(self) -> asyncio.Future | None:
        """
        Loop over the various connection methods. Only wait minimally before
        retries within the same loop. If everything fails, the connection
        is rescheduled for later.
        """
        if self._connect_loop_wait > 0:
            self.event('reconnect_delay', self._connect_loop_wait)
            await asyncio.sleep(self._connect_loop_wait)
        if self.custom_address:
            host, port = self.custom_address
            records = [('', self.default_domain, host, port)]
        else:
            records = await self.get_dns_records(self.default_domain)
        if not records: # No DNS records
            records = [
                ('', self.default_domain, self.default_domain, self.default_port),
            ]
        success = False
        for record in records:
            if success:
                break
            service, host, address, dns_port = record
            tls = service in self.tls_services
            server_hostname = None
            if tls:
                server_hostname = self.default_domain
            try:
                if not service:
                    fake_services = []
                    if self.enable_direct_tls:
                        fake_services.extend(list(self.tls_services))
                    if self.enable_starttls:
                        fake_services.extend(list(self.starttls_services))
                    if self.enable_plaintext:
                        fake_services.extend([''])
                    for service in fake_services:
                        tls, server_hostname = (False, None)
                        if service in self.tls_services:
                            tls, server_hostname = True, self.default_domain
                        success = await self._attempt_connection(
                            address, dns_port, tls, server_hostname,
                        )
                        if success:
                            break
                else:
                    success = await self._attempt_connection(
                        address, dns_port, tls, server_hostname,
                    )
            except Exception as exc:
                log.error('Unhandled exception during connect(): %s',
                          exc, exc_info=True)
        if success:
            return None
        if self._current_connection_attempt is not None:
            return self.reschedule_connection_attempt()
        return None

    async def _attempt_connection(self, host: str, port: int, tls: bool,
                                  server_hostname: str | None) -> bool:
        """Try to connect to a remote server."""
        self.event_when_connected = "connected"
        self._connect_loop_wait += 1
        ssl_context: ssl.SSLContext | None = None
        if tls:
            ssl_context = self.get_ssl_context()

        if self._current_connection_attempt is None:
            return False
        try:
            await self.loop.create_connection(lambda: self,
                                              host, port,
                                              ssl=ssl_context,
                                              server_hostname=server_hostname)
            self._connect_loop_wait = 0
            return True
        except Socket.gaierror:
            self.event('connection_failed',
                       'No DNS record available for %s' % self.default_domain)
            return False
        except OSError as e:
            log.debug('Connection failed: %s', e)
            self.event("connection_failed", e)
            return False

    def init_parser(self) -> None:
        """init the XML parser. The parser must always be reset for each new
        connexion
        """
        self.xml_depth = 0
        self.xml_root = None
        self.parser = ET.XMLPullParser(("start", "end"))

    def connection_made(self, transport: BaseTransport, send_event: bool = True) -> None:
        """Called when the TCP connection has been established with the server
        """
        if send_event:
            self.event(self.event_when_connected)
        self.transport = cast(Transport, transport)
        if self.transport is None:
            raise ValueError("Transport cannot be none")
        self.socket = self.transport.get_extra_info(
            "ssl_object",
            default=self.transport.get_extra_info("socket")
        )
        ssl_object = transport.get_extra_info(
            "ssl_object",
            default=None,
        )
        if ssl_object is not None:
            der_cert = ssl_object.getpeercert(True)
            pem_cert = ssl.DER_cert_to_PEM_cert(der_cert)
            self.event('ssl_cert', pem_cert)
            if self._current_connection_attempt is None:
                # Connection attempt aborted
                return
            self.event('tls_success')
        self._current_connection_attempt = None
        self.init_parser()
        self.send_raw(self.stream_header)

    def data_received(self, data: bytes | str) -> None:
        """Called when incoming data is received on the socket.

        We feed that data to the parser and the see if this produced any XML
        event.  This could trigger one or more event (a stanza is received,
        the stream is opened, etc).
        """
        if self.parser is None:
            log.warning('Received data before the connection is established: %r',
                        data)
            return
        self.parser.feed(data)
        try:
            for event, xml in self.parser.read_events():  # type:ignore[misc]
                xml = cast(ET.Element, xml)
                if event == 'start':
                    if self.xml_depth == 0:
                        # We have received the start of the root element.
                        self.xml_root = xml
                        log.debug('RECV: %s', tostring(self.xml_root,
                                                       xmlns=self.default_ns,
                                                       stream=self,
                                                       top_level=True,
                                                       open_only=True))
                        self.start_stream_handler(self.xml_root)  # type:ignore
                    self.xml_depth += 1
                if event == 'end':
                    self.xml_depth -= 1
                    if self.xml_depth == 0:
                        # The stream's root element has closed,
                        # terminating the stream.
                        log.debug("End of stream received")
                        self.disconnect_reason = "End of stream"
                        self.abort()
                        return
                    elif self.xml_depth == 1:
                        # A stanza is an XML element that is a direct child of
                        # the root element, hence the check of depth == 1
                        self._spawn_event(xml)
                        if self.xml_root is not None:
                            # Keep the root element empty of children to
                            # save on memory use.
                            self.xml_root.clear()
        except ET.ParseError:
            log.error('Parse error: %r', data)

            # Due to cyclic dependencies, this can’t be imported at the module
            # level.
            from slixmpp.stanza.stream_error import StreamError
            error = StreamError()
            error['condition'] = 'not-well-formed'
            error['text'] = 'Server sent: %r' % data
            self.send(error)
            self.disconnect()

    def is_connecting(self) -> bool:
        return self._current_connection_attempt is not None

    def is_connected(self) -> bool:
        return self.transport is not None

    def eof_received(self) -> None:
        """When the TCP connection is properly closed by the remote end
        """
        self.event("eof_received")

    def connection_lost(self, exception: BaseException | None) -> None:
        """On any kind of disconnection, initiated by us or not.  This signals the
        closure of the TCP connection
        """
        log.info("connection_lost: %s", (exception,))
        # All these objects are associated with one TCP connection.  Since
        # we are not connected anymore, destroy them
        self.parser = None
        self.transport = None
        self.socket = None
        # Fire the events after cleanup
        if self.end_session_on_disconnect:
            self._reset_sendq()
            self.event('session_end')
        self._set_disconnected_future()
        self.event("disconnected", self.disconnect_reason or exception)

    def reschedule_connection_attempt(self) -> asyncio.Future | None:
        """
        Increase the exponential back-off and initiate another background
        _connect_loop call to connect to the server.

        :returns: A future on the next scheduled connection attempt.
        """
        # abort if there is no ongoing connection attempt
        if self._current_connection_attempt is None:
            return None
        self._connect_loop_wait = min(300, self._connect_loop_wait * 2 + 1)
        self._current_connection_attempt = asyncio.ensure_future(
            self._connect_loop(),
            loop=self.loop,
        )
        return self._current_connection_attempt

    def cancel_connection_attempt(self) -> None:
        """
        Immediately cancel the current create_connection() Future.
        This is useful when a client using slixmpp tries to connect
        on flaky networks, where sometimes a connection just gets lost
        and it needs to reconnect while the attempt is still ongoing.
        """
        if self._current_connection_attempt:
            self._current_connection_attempt.cancel()
            self._current_connection_attempt = None

    def disconnect(self, wait: float | int = 2.0, reason: str | None = None, ignore_send_queue: bool = False) -> Future:
        """Close the XML stream and wait for an acknowldgement from the server for
        at most `wait` seconds.  After the given number of seconds has
        passed without a response from the server, or when the server
        successfully responds with a closure of its own stream, abort() is
        called. If wait is 0.0, this will call abort() directly without closing
        the stream.

        Does nothing but trigger the disconnected event if we are not connected.

        :param wait: Time to wait for a response from the server.
        :param reason: An optional reason for the disconnect.
        :param ignore_send_queue: Boolean to toggle if we want to ignore
                                  the in-flight stanzas and disconnect immediately.
        :return: A future that ends when all code involved in the disconnect has ended
        """
        # Compat: docs/getting_started/sendlogout.rst has been promoting
        # `disconnect(wait=True)` for ages. This doesn't mean anything to the
        # schedule call below. It would fortunately be converted to `1` later
        # down the call chain. Praise the implicit casts lord.
        if wait is True:
            wait = 2.0

        if self.transport:
            self.disconnect_reason = reason
            if self.waiting_queue.empty() or ignore_send_queue:
                self.cancel_connection_attempt()
                return asyncio.ensure_future(
                    self._end_stream_wait(wait, reason=reason),
                    loop=self.loop,
                )
            else:
                return asyncio.ensure_future(
                    self._consume_send_queue_before_disconnecting(reason, wait),
                    loop=self.loop,
                )
        else:
            self._set_disconnected_future()
            self.event("disconnected", reason)
            future: Future = Future(loop=self.loop)
            future.set_result(None)
            return future

    async def _consume_send_queue_before_disconnecting(self, reason: str | None, wait: float) -> None:
        """Wait until the send queue is empty before disconnecting"""
        try:
            await asyncio.wait_for(
                self.waiting_queue.join(),
                wait,
            )
        except asyncio.TimeoutError:
            wait = 0 # we already consumed the timeout
        self.disconnect_reason = reason
        await self._end_stream_wait(wait)

    async def _end_stream_wait(self, wait: int | float = 2, reason: str | None = None) -> None:
        """
        Run abort() if we do not received the disconnected event
        after a waiting time.

        :param wait: The waiting time (defaults to 2)
        """
        try:
            self.send_raw(self.stream_footer)
            await self.wait_until('disconnected', wait)
        except asyncio.TimeoutError:
            self.abort()
        except NotConnectedError:
            # We are not connected when sending the end of stream
            # that means the disconnect has already been handled
            pass

    def abort(self) -> None:
        """
        Forcibly close the connection
        """
        if self.transport:
            self.cancel_connection_attempt()
            self.transport.close()
            self.transport.abort()
            self.event("killed")

    def reconnect(self, wait: int | float = 2.0, reason: str = "Reconnecting") -> None:
        """Calls disconnect(), and once we are disconnected (after the timeout, or
        when the server acknowledgement is received), call connect()
        """
        log.debug("reconnecting...")
        async def handler(event: Any) -> None:
            # We yield here to allow synchronous handlers to work first
            await asyncio.sleep(0)
            host, port = None, None
            if self.custom_address:
                host, port = self.custom_address
            self.connect(host=host, port=port)
        self.add_event_handler('disconnected', handler, disposable=True)
        self.disconnect(wait, reason)

    def configure_socket(self) -> None:
        """Set timeout and other options for self.socket.

        Meant to be overridden.
        """
        pass

    def configure_dns(self, resolver: Any, domain: str | None = None, port: int | None = None) -> None:
        """
        Configure and set options for a :class:`~dns.resolver.Resolver`
        instance, and other DNS related tasks. For example, you
        can also check :meth:`~socket.socket.getaddrinfo` to see
        if you need to call out to ``libresolv.so.2`` to
        run ``res_init()``.

        Meant to be overridden.

        :param resolver: A :class:`~dns.resolver.Resolver` instance
                         or ``None`` if ``dnspython`` is not installed.
        :param domain: The initial domain under consideration.
        :param port: The initial port under consideration.
        """
        pass

    def get_ssl_context(self) -> ssl.SSLContext:
        """
        Get SSL context.
        """
        if self.ciphers is not None:
            self.ssl_context.set_ciphers(self.ciphers)
        if self.keyfile and self.certfile:
            try:
                self.ssl_context.load_cert_chain(self.certfile, self.keyfile)
            except (ssl.SSLError, OSError):
                log.debug('Error loading the cert chain:', exc_info=True)
            else:
                log.debug('Loaded cert file %s and key file %s',
                          self.certfile, self.keyfile)
        if self.ca_certs is not None:
            ca_cert: Path | None = None
            # XXX: Compat before d733c54518.
            if isinstance(self.ca_certs, str):
                self.ca_certs = Path(self.ca_certs)
            if isinstance(self.ca_certs, Path):
                if self.ca_certs.is_file():
                    ca_cert = self.ca_certs
            else:
                for bundle in self.ca_certs:
                    if bundle.is_file():
                        ca_cert = bundle
                        break
            if ca_cert is None and \
               isinstance(self.ca_certs, (Path, collections.abc.Iterable)):
                raise InvalidCABundle(self.ca_certs)

            self.ssl_context.verify_mode = ssl.CERT_REQUIRED
            self.ssl_context.load_verify_locations(cafile=ca_cert)

        return self.ssl_context

    async def start_tls(self) -> bool:
        """Perform handshakes for TLS.

        If the handshake is successful, the XML stream will need
        to be restarted.
        """
        if self.transport is None:
            raise ValueError("Transport should not be None")
        ssl_context = self.get_ssl_context()
        try:
            self._current_connection_attempt = asyncio.ensure_future(
                self.loop.start_tls(self.transport,
                                    self, ssl_context,
                                    server_hostname=self.default_domain),
                loop=self.loop,
            )
            transp = await self._current_connection_attempt
        except ssl.SSLError as e:
            log.debug('SSL: Unable to connect', exc_info=True)
            log.error('CERT: Invalid certificate trust chain.')
            if not self.event_handled('ssl_invalid_chain'):
                self.disconnect()
            else:
                self.event('ssl_invalid_chain', e)
            return False
        except OSError:
            log.debug("Connection error:", exc_info=True)
            self.disconnect()
            return False
        if transp is None:
            raise Exception("Transport should not be none")
        # If we use the builtin start_tls, the connection_made() protocol
        # method is not called automatically
        if hasattr(self.loop, 'start_tls'):
            self.connection_made(transp, send_event=False)
        return True

    def _start_keepalive(self, event: Any) -> None:
        """Begin sending whitespace periodically to keep the connection alive.

        May be disabled by setting::

            self.whitespace_keepalive = False

        The keepalive interval can be set using::

            self.whitespace_keepalive_interval = 300
        """
        self.schedule('Whitespace Keepalive',
                      self.whitespace_keepalive_interval,
                      self.send_raw,
                      args=(' ',),
                      repeat=True)

    def _remove_schedules(self, event: Any) -> None:
        """Remove some schedules that become pointless when disconnected"""
        self.cancel_schedule('Whitespace Keepalive')

    def start_stream_handler(self, xml: ET.Element) -> None:
        """Perform any initialization actions, such as handshakes,
        once the stream header has been sent.

        Meant to be overridden.
        """
        pass

    def register_stanza(self, stanza_class: type[StanzaBase]) -> None:
        """Add a stanza object class as a known root stanza.

        A root stanza is one that appears as a direct child of the stream's
        root element.

        Stanzas that appear as substanzas of a root stanza do not need to
        be registered here. That is done using register_stanza_plugin() from
        slixmpp.xmlstream.stanzabase.

        Stanzas that are not registered will not be converted into
        stanza objects, but may still be processed using handlers and
        matchers.

        :param stanza_class: The top-level stanza object's class.
        """
        self.__root_stanza.append(stanza_class)

    def remove_stanza(self, stanza_class: type[StanzaBase]) -> None:
        """Remove a stanza from being a known root stanza.

        A root stanza is one that appears as a direct child of the stream's
        root element.

        Stanzas that are not registered will not be converted into
        stanza objects, but may still be processed using handlers and
        matchers.
        """
        self.__root_stanza.remove(stanza_class)

    def add_filter(self, mode: FilterString, handler: Callable[[StanzaBase], StanzaBase | None], order: int | None = None) -> None:
        """Add a filter for incoming or outgoing stanzas.

        These filters are applied before incoming stanzas are
        passed to any handlers, and before outgoing stanzas
        are put in the send queue.

        Each filter must accept a single stanza, and return
        either a stanza or ``None``. If the filter returns
        ``None``, then the stanza will be dropped from being
        processed for events or from being sent.

        :param mode: One of ``'in'`` or ``'out'``.
        :param handler: The filter function.
        :param int order: The position to insert the filter in
                          the list of active filters.
        """
        if order:
            self.__filters[mode].insert(order, handler)
        else:
            self.__filters[mode].append(handler)

    def del_filter(self, mode: str, handler: Callable[[StanzaBase], StanzaBase | None]) -> None:
        """Remove an incoming or outgoing filter."""
        self.__filters[mode].remove(handler)

    def register_handler(self, handler: BaseHandler, before: BaseHandler | None = None, after: BaseHandler | None = None) -> None:
        """Add a stream event handler that will be executed when a matching
        stanza is received.

        :param handler:
                The :class:`~slixmpp.xmlstream.handler.base.BaseHandler`
                derived object to execute.
        """
        if handler.stream is None:
            self.__handlers.append(handler)
            handler.stream = weakref.ref(self)

    def remove_handler(self, name: str) -> bool:
        """Remove any stream event handlers with the given name.

        :param name: The name of the handler.
        """
        idx = 0
        for handler in self.__handlers:
            if handler.name == name:
                self.__handlers.pop(idx)
                return True
            idx += 1
        return False

    async def get_dns_records(self, domain: str, port: int | None = None) -> list[tuple[str, str, str, int]]:
        """Get the DNS records for a domain.

        :param domain: The domain in question.
        :param port: If the results don't include a port, use this one.
        """
        if port is None:
            port = self.default_port

        resolver = default_resolver(loop=self.loop)
        self.configure_dns(resolver, domain=domain, port=port)

        services = []
        if self.enable_direct_tls:
            services.extend(list(self.tls_services))
        if self.enable_starttls:
            services.extend(list(self.starttls_services))
        result = await resolve(domain, port,
                               services=services,
                               resolver=resolver,
                               use_ipv6=self.use_ipv6,
                               use_aiodns=self.use_aiodns,
                               loop=self.loop)
        return result

    def add_event_handler(self, name: str, pointer: Callable[..., Any], disposable: bool = False) -> None:
        """Add a custom event handler that will be executed whenever
        its event is manually triggered.

        :param name: The name of the event that will trigger
                     this handler.
        :param pointer: The function to execute.
        :param disposable: If set to ``True``, the handler will be
                           discarded after one use. Defaults to ``False``.
        """
        if name not in self.__event_handlers:
            self.__event_handlers[name] = []
        self.__event_handlers[name].append((pointer, disposable))

    def del_event_handler(self, name: str, pointer: Callable[..., Any]) -> None:
        """Remove a function as a handler for an event.

        :param name: The name of the event.
        :param pointer: The function to remove as a handler.
        """
        if name not in self.__event_handlers:
            return

        # Need to keep handlers that do not use
        # the given function pointer
        def filter_pointers(handler: tuple[Callable[..., Any], bool]) -> bool:
            return handler[0] != pointer

        self.__event_handlers[name] = list(filter(
            filter_pointers,
            self.__event_handlers[name]))

    def event_handled(self, name: str) -> int:
        """Returns the number of registered handlers for an event.

        :param name: The name of the event to check.
        """
        return len(self.__event_handlers.get(name, []))

    async def event_async(self, name: str, data: Any = {}) -> None:
        """Manually trigger a custom event, but await coroutines immediately.

        This event generator should only be called in situations when
        in-order processing of events is important, such as features
        handling.

        :param name: The name of the event to trigger.
        :param data: Data that will be passed to each event handler.
                     Defaults to an empty dictionary, but is usually
                     a stanza object.
        """
        handlers = self.__event_handlers.get(name, [])[:]
        for handler in handlers:
            handler_callback, disposable = handler
            if disposable:
                # If the handler is disposable, we will go ahead and
                # remove it now instead of waiting for it to be
                # processed in the queue.
                try:
                    self.__event_handlers[name].remove(handler)
                except ValueError:
                    pass
            # If the callback is a coroutine, schedule it instead of
            # running it directly
            if iscoroutinefunction(handler_callback):
                try:
                    await handler_callback(data)
                except Exception as exc:
                    self.exception(exc)
            else:
                try:
                    handler_callback(data)
                except Exception as e:
                    self.exception(e)

    def event(self, name: str, data: Any = {}) -> None:
        """Manually trigger a custom event.
        Coroutine handlers are wrapped into a future and sent into the
        event loop for their execution, and not awaited.

        :param name: The name of the event to trigger.
        :param data: Data that will be passed to each event handler.
                     Defaults to an empty dictionary, but is usually
                     a stanza object.
        """
        log.debug("Event triggered: %s", name)

        handlers = self.__event_handlers.get(name, [])[:]
        for handler in handlers:
            handler_callback, disposable = handler
            old_exception = getattr(data, 'exception', None)

            # If the callback is a coroutine, schedule it instead of
            # running it directly
            if iscoroutinefunction(handler_callback):
                async def handler_callback_routine(cb: Callable[[ElementBase], Any]) -> None:
                    try:
                        await cb(data)
                    except Exception as e:
                        if old_exception:
                            old_exception(e)
                        else:
                            self.exception(e)
                asyncio.ensure_future(
                    handler_callback_routine(handler_callback),
                    loop=self.loop,
                )
            else:
                try:
                    handler_callback(data)
                except Exception as e:
                    if old_exception:
                        old_exception(e)
                    else:
                        self.exception(e)
            if disposable:
                # If the handler is disposable, we will go ahead and
                # remove it now instead of waiting for it to be
                # processed in the queue.
                try:
                    self.__event_handlers[name].remove(handler)
                except ValueError:
                    pass

    def schedule(self, name: str, seconds: int, callback: Callable[..., None],
            args: tuple[Any, ...] = tuple(),
            kwargs: dict[Any, Any] = {}, repeat: bool = False) -> None:
        """Schedule a callback function to execute after a given delay.

        :param name: A unique name for the scheduled callback.
        :param seconds: The time in seconds to wait before executing.
        :param callback: A pointer to the function to execute.
        :param args: A tuple of arguments to pass to the function.
        :param kwargs: A dictionary of keyword arguments to pass to
                       the function.
        :param repeat: Flag indicating if the scheduled event should
                       be reset and repeat after executing.
        """
        if name in self.scheduled_events:
            raise ValueError(
                "There is already a scheduled event of name: %s" % name)
        if seconds is None:
            seconds = RESPONSE_TIMEOUT
        cb = functools.partial(callback, *args, **kwargs)
        if repeat:
            handle = self.loop.call_later(seconds, self._execute_and_reschedule,
                                          name, cb, seconds)
        else:
            handle = self.loop.call_later(seconds, self._execute_and_unschedule,
                                          name, cb)
        # Save that handle, so we can just cancel this scheduled event by
        # canceling scheduled_events[name]
        self.scheduled_events[name] = handle

    def cancel_schedule(self, name: str) -> None:
        try:
            handle = self.scheduled_events.pop(name)
            handle.cancel()
        except KeyError:
            log.debug("Tried to cancel unscheduled event: %s" % (name,))

    def _safe_cb_run(self, name: str, cb: Callable[[], None]) -> None:
        log.debug('Scheduled event: %s', name)
        try:
            cb()
        except Exception as e:
            self.exception(e)

    def _execute_and_reschedule(self, name: str, cb: Callable[[], None], seconds: int) -> None:
        """Simple method that calls the given callback, and then schedule itself to
        be called after the given number of seconds.
        """
        self._safe_cb_run(name, cb)
        handle = self.loop.call_later(seconds, self._execute_and_reschedule,
                                      name, cb, seconds)
        self.scheduled_events[name] = handle

    def _execute_and_unschedule(self, name: str, cb: Callable[[], None]) -> None:
        """
        Execute the callback and remove the handler for it.
        """
        self._safe_cb_run(name, cb)
        # workaround for specific events which unschedule themselves
        if name in self.scheduled_events:
            del self.scheduled_events[name]

    def incoming_filter(self, xml: ET.Element) -> ET.Element:
        """Filter incoming XML objects before they are processed.

        Possible uses include remapping namespaces, or correcting elements
        from sources with incorrect behavior.

        Meant to be overridden.
        """
        return xml

    def _reset_sendq(self) -> None:
        """Clear sending tasks on session end"""
        # Cancel all pending slow send tasks
        log.debug('Cancelling %d slow send tasks', len(self.__slow_tasks))
        for slow_task in self.__slow_tasks:
            slow_task.cancel()
        self.__slow_tasks.clear()
        # Purge pending stanzas
        while not self.waiting_queue.empty():
            discarded = self.waiting_queue.get_nowait()
            log.debug('Discarded stanza: %s', discarded)

    async def _continue_slow_send(
            self,
            task: asyncio.Task,
            already_used: set[Filter]
    ) -> None:
        """
        Used when an item in the send queue has taken too long to process.

        This is away from the send queue and can take as much time as needed.
        :param asyncio.Task task: the Task wrapping the coroutine
        :param set already_used: Filters already used on this outgoing stanza
        """
        data = await task
        self.__slow_tasks.remove(task)
        if data is None:
            return
        for filter in self.__filters['out'][:]:
            if filter in already_used:
                continue
            if iscoroutinefunction(filter):
                data = await filter(data)  # type: ignore
            else:
                filter = cast(SyncFilter, filter)
                data = filter(data)
            if data is None:
                return

        if isinstance(data, StanzaBase):
            for filter in self.__filters['out_sync']:
                filter = cast(SyncFilter, filter)
                data = filter(data)
                if data is None:
                    return
            str_data = tostring(data.xml, xmlns=self.default_ns,
                                stream=self, top_level=True)
            self.send_raw(str_data)
        else:
            self.send_raw(data)

    async def run_filters(self) -> None:
        """
        Background loop that processes stanzas to send.
        """
        while True:
            data: StanzaBase | str | None
            (data, use_filters) = await self.waiting_queue.get()
            try:
                if isinstance(data, StanzaBase):
                    if use_filters:
                        already_run_filters = set()
                        for filter in self.__filters['out']:
                            already_run_filters.add(filter)
                            if iscoroutinefunction(filter):
                                filter = cast(AsyncFilter, filter)
                                task = asyncio.create_task(filter(data))  # type:ignore
                                completed, pending = await wait(
                                    {task},
                                    timeout=1,
                                )
                                if pending:
                                    self.__slow_tasks.append(task)
                                    asyncio.ensure_future(
                                        self._continue_slow_send(
                                            task,
                                            already_run_filters
                                        ),
                                        loop=self.loop,
                                    )
                                    raise ContinueQueue(
                                        "Slow coroutine, rescheduling filters"
                                    )
                                data = task.result()
                            elif isinstance(data, StanzaBase):
                                filter = cast(SyncFilter, filter)
                                data = filter(data)
                            if data is None:
                                raise ContinueQueue('Empty stanza')

                if isinstance(data, StanzaBase):
                    if use_filters:
                        for filter in self.__filters['out_sync']:
                            filter = cast(SyncFilter, filter)
                            data = filter(data)
                            if data is None:
                                raise ContinueQueue('Empty stanza')
                    if isinstance(data, StanzaBase):
                        str_data = tostring(data.xml, xmlns=self.default_ns,
                                            stream=self, top_level=True)
                    else:
                        str_data = data
                    self.send_raw(str_data)
                elif isinstance(data, (str, bytes)):
                    self.send_raw(data)
            except ContinueQueue as exc:
                log.debug('Stanza in send queue not sent: %s', exc)
            except asyncio.CancelledError:
                log.debug('Send coroutine received cancel(), stopping')
                return
            except Exception:
                log.error('Exception raised in send queue:', exc_info=True)
            self.waiting_queue.task_done()

    def send(self, data: StanzaBase | str, use_filters: bool = True) -> None:
        """A wrapper for :meth:`send_raw()` for sending stanza objects.

        :param data: The :class:`~slixmpp.xmlstream.stanzabase.StanzaBase`
                     stanza to send on the stream.
        :param bool use_filters: Indicates if outgoing filters should be
                                 applied to the given stanza data. Disabling
                                 filters is useful when resending stanzas.
                                 Defaults to ``True``.
        """
        # When not connected, allow features/starttls/etc to go through
        # but not stanzas or arbitrary payloads.
        if not self._always_send_everything and not self._session_started:
            # Avoid circular imports
            from slixmpp.stanza.rootstanza import RootStanza
            from slixmpp.stanza import Iq, Handshake

            passthrough = False
            if isinstance(data, Iq):
                if data.get_plugin('bind', check=True):
                    passthrough = True
                elif data.get_plugin('session', check=True):
                    passthrough = True
                elif data.get_plugin('register', check=True):
                    passthrough = True
            elif isinstance(data, Handshake):
                passthrough = True

            if isinstance(data, (RootStanza, str)) and not passthrough:
                self.__queued_stanzas.append((data, use_filters))
                log.debug('NOT SENT: %s %s', type(data), data)
                self.event('stanza_not_sent', data)
                return
        self.waiting_queue.put_nowait((data, use_filters))

    def send_xml(self, data: ET.Element) -> None:
        """Send an XML object on the stream

        :param data: The :class:`~xml.etree.ElementTree.Element` XML object
                     to send on the stream.
        """
        self.send(tostring(data))

    def send_raw(self, data: str | bytes) -> None:
        """Send raw data across the stream.

        :param string data: Any bytes or utf-8 string value.
        """
        log.debug("SEND: %s", data)
        if not self.transport:
            raise NotConnectedError()
        if isinstance(data, str):
            data = data.encode('utf-8')
        self.transport.write(data)

    def _build_stanza(self, xml: ET.Element,
                      default_ns: str | None = None) -> StanzaBase:
        """Create a stanza object from a given XML object.

        If a specialized stanza type is not found for the XML, then
        a generic :class:`~slixmpp.xmlstream.stanzabase.StanzaBase`
        stanza will be returned.

        :param xml: The :class:`~xml.etree.ElementTree.Element` XML object
                    to convert into a stanza object.
        :param default_ns: Optional default namespace to use instead of the
                           stream's current default namespace.
        """
        if default_ns is None:
            default_ns = self.default_ns
        stanza_type = StanzaBase
        for stanza_class in self.__root_stanza:
            if xml.tag == "{%s}%s" % (default_ns, stanza_class.name) or \
               xml.tag == stanza_class.tag_name():
                stanza_type = stanza_class
                break
        stanza = stanza_type(self, xml, recv=True)
        if stanza['lang'] is None and self.peer_default_lang:
            stanza['lang'] = self.peer_default_lang
        return stanza

    def _spawn_event(self, xml: ET.Element) -> None:
        """
        Analyze incoming XML stanzas and convert them into stanza
        objects if applicable and queue stream events to be processed
        by matching handlers.

        :param xml: The :class:`~slixmpp.xmlstream.stanzabase.ElementBase`
                    stanza to analyze.
        """
        # Apply any preprocessing filters.
        xml = self.incoming_filter(xml)

        # Convert the raw XML object into a stanza object. If no registered
        # stanza type applies, a generic StanzaBase stanza will be used.
        try:
            stanza: StanzaBase | None = self._build_stanza(xml)
        except Exception as exc:
            log.exception("Unable to parse stanza: %s,\n%s", exc, xml)
            stanza = None
        for filter in self.__filters['in']:
            if stanza is not None:
                filter = cast(SyncFilter, filter)
                stanza = filter(stanza)
        if stanza is None:
            return

        log.debug("RECV: %s", stanza)

        # Match the stanza against registered handlers. Handlers marked
        # to run "in stream" will be executed immediately; the rest will
        # be queued.
        handled = False
        matched_handlers = [h for h in self.__handlers if h.match(stanza)]
        for handler in matched_handlers:
            handler.prerun(stanza)
            try:
                handler.run(stanza)
            except Exception as e:
                stanza.exception(e)
            if handler.check_delete():
                self.__handlers.remove(handler)
            handled = True

        # Some stanzas require responses, such as Iq queries. A default
        # handler will be executed immediately for this case.
        if not handled:
            stanza.unhandled()

    def exception(self, exception: Exception) -> None:
        """Process an unknown exception.

        Meant to be overridden.

        :param exception: An unhandled exception object.
        """
        pass

    async def wait_until(self, event: str, timeout: int | float = 30) -> Any:
        """Utility method to wake on the next firing of an event.
        (Registers a disposable handler on it)

        :param str event: Event to wait on.
        :param int timeout: Timeout
        :raises: :class:`asyncio.TimeoutError` when the timeout is reached
        """
        fut: Future = asyncio.Future(loop=self.loop)

        def result_handler(event_data: Any) -> None:
            if not fut.done():
                fut.set_result(event_data)
            else:
                log.debug(
                    "Future registered on event '%s' was already done",
                    event
                )

        self.add_event_handler(
            event,
            result_handler,
            disposable=True,
        )
        return await asyncio.wait_for(fut, timeout)

    @contextmanager
    def event_handler(self, event: str, handler: Callable[..., Any]) -> Generator[None, None, None]:
        """
        Context manager that adds then removes an event handler.
        """
        self.add_event_handler(event, handler)
        try:
            yield
        except Exception:
            raise
        finally:
            self.del_event_handler(event, handler)

    def wrap(self, coroutine: Awaitable[T]) -> Future[T]:
        """Make a Future out of a coroutine or another awaitable
        with the current loop.

        :param coroutine: The coroutine to wrap.
        """
        return asyncio.ensure_future(
            coroutine,
            loop=self.loop,
        )

    def __del__(self) -> None:
        if self._run_out_filters is not None:
            self._run_out_filters.cancel()



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp import JID
from slixmpp.plugins import BasePlugin
from slixmpp.stanza import Presence
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream.handler import Callback

from slixmpp.plugins.xep_0437 import stanza


class XEP_0437(BasePlugin):
    """
    XEP-0437: Room Activity Indicators plugin


    Defines two events:

    - ``room_activity_bare``: useful for components receiving a subscription request.
    - ``room_activity``: when receiving a room activity.
    """
    name = 'xep_0437'
    description = 'XEP-0437: Room Activity Indicators'
    stanza = stanza
    namespace = stanza.NS

    def plugin_init(self):
        stanza.register_plugins()
        self.xmpp.register_handler(Callback(
            'RAI Received',
            StanzaPath("presence/rai"),
            self._handle_rai,
        ))
        self.xmpp.register_handler(Callback(
            'RAI Activity Received',
            StanzaPath("presence/rai/activity"),
            self._handle_rai_activity,
        ))

    def plugin_end(self):
        self.xmpp.remove_handler('RAI received')
        self.xmpp.remove_handler('RAI Activity received')

    def _handle_rai(self, presence: Presence):
        self.xmpp.event('room_activity_bare', presence)

    def _handle_rai_activity(self, presence: Presence):
        self.xmpp.event('room_activity', presence)

    def subscribe(self, service: JID, *,
                  pfrom: JID | None = None):
        """
        Subscribe to room activity on a MUC service.

        :param JID service: MUC service
        :param pfrom: JID the subscribe request is sent from (for components).
        """
        pres = self.xmpp.make_presence(pto=service, pfrom=pfrom)
        pres.enable('rai')
        pres.send()

    def unsubscribe(self, service: JID, *,
                    pfrom: JID | None = None):
        """
        Unsubscribe from room activity on a MUC service.

        :param service: MUC service.
        :param pfrom: JID the unsubscribe request is sent from (for components).
        """
        pres = self.xmpp.make_presence(
            pto=service, pfrom=pfrom, ptype='unavailable',
        )
        pres.send()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
__all__ = [
    'feature_starttls',
    'feature_mechanisms',
    'feature_bind',
    'feature_session',
    'feature_rosterver',
    'feature_preapproval',
    'feature_limits',
]




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.plugins import BasePlugin, register_plugin


class XEP_0106(BasePlugin):

    name = 'xep_0106'
    description = 'XEP-0106: JID Escaping'
    dependencies = {'xep_0030'}

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature(feature='jid\\20escaping')

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature='jid\\20escaping')


register_plugin(XEP_0106)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin
from .retraction import XEP_0424

register_plugin(XEP_0424)

__all__ = ['XEP_0424']



import warnings
from typing import Iterable

from slixmpp import Presence
from slixmpp.types import HatTuple
from slixmpp.xmlstream import ElementBase, register_stanza_plugin

NS = 'urn:xmpp:hats:0'


class Hats(ElementBase):
    """
    Hats element, container for multiple hats:

    .. code-block::xml


      <hats xmlns='urn:xmpp:hats:0'>
        <hat title='Host' uri='http://schemas.example.com/hats#host' xml:lang='en-us'>
            <badge xmlns="urn:example:badges" fgcolor="#000000" bgcolor="#58C5BA"/>
        </hat>
        <hat title='Presenter' uri='http://schemas.example.com/hats#presenter' xml:lang='en-us'>
            <badge xmlns="urn:example:badges" fgcolor="#000000" bgcolor="#EC0524"/>
        </hat>
      </hats>

    """

    name = 'hats'
    namespace = NS
    plugin_attrib = 'hats'

    def add_hats(self, hats: Iterable[HatTuple | tuple[str, str, float]]) -> None:
        for uri, title, hue in hats:
            hat = Hat()
            hat["uri"] = uri
            hat["title"] = title
            if hue is not None:
                hat["hue"] = hue
            self.append(hat)


class Hat(ElementBase):
    """
    Hat element, has a title and url, may contain arbitrary sub-elements.

    .. code-block::xml

        <hat title='Host' uri='http://schemas.example.com/hats#host' xml:lang='en-us'>
            <badge xmlns="urn:example:badges" fgcolor="#000000" bgcolor="#58C5BA"/>
        </hat>

    """
    name = 'hat'
    plugin_attrib = 'hat'
    namespace = NS
    interfaces = {'title', 'uri', 'hue'}
    plugin_multi_attrib = "hats"

    def set_hue(self, hue: float) -> None:
        self._set_attr("hue", str(hue))

    def get_hue(self) -> float | None:
        hue = self._get_attr("hue",)
        try:
            return None if hue == "" else float(hue)
        except ValueError:
            warnings.warn(f"Not a valid hue value: {hue}")
            return None


def register_plugin() -> None:
    register_stanza_plugin(Hats, Hat, iterable=True)
    register_stanza_plugin(Presence, Hats)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class Signed(ElementBase):
    name = 'x'
    namespace = 'jabber:x:signed'
    plugin_attrib = 'signed'
    interfaces = {'signed'}
    is_extension = True

    def set_signed(self, value):
        parent = self.parent()
        xmpp = parent.stream
        data = xmpp['xep_0027'].sign(value, parent['from'])
        if data:
            self.xml.text = data
        else:
            del parent['signed']

    def get_signed(self):
        return self.xml.text


class Encrypted(ElementBase):
    name = 'x'
    namespace = 'jabber:x:encrypted'
    plugin_attrib = 'encrypted'
    interfaces = {'encrypted'}
    is_extension = True

    def set_encrypted(self, value):
        parent = self.parent()
        xmpp = parent.stream
        data = xmpp['xep_0027'].encrypt(value, parent['to'])
        if data:
            self.xml.text = data
        else:
            del parent['encrypted']

    def get_encrypted(self):
        parent = self.parent()
        xmpp = parent.stream
        if self.xml.text:
            return xmpp['xep_0027'].decrypt(self.xml.text, parent['to'])
        return None




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ET, ElementBase


class Headers(ElementBase):
    name = 'headers'
    namespace = 'http://jabber.org/protocol/shim'
    plugin_attrib = 'headers'
    interfaces = {'headers'}
    is_extension = True

    def get_headers(self):
        result = {}
        headers = self.xml.findall('{%s}header' % self.namespace)
        for header in headers:
            name = header.attrib.get('name', '')
            value = header.text
            if name in result:
                if not isinstance(result[name], set):
                    result[name] = [result[name]]
                else:
                    result[name] = []
                result[name].add(value)
            else:
                result[name] = value
        return result

    def set_headers(self, values):
        self.del_headers()
        for name in values:
            vals = values[name]
            if not isinstance(vals, (list, set)):
                vals = [values[name]]
            for value in vals:
                header = ET.Element('{%s}header' % self.namespace)
                header.attrib['name'] = name
                header.text = value
                self.xml.append(header)

    def del_headers(self):
        headers = self.xml.findall('{%s}header' % self.namespace)
        for header in headers:
            self.xml.remove(header)



'''Pure asyncio implementation of RFC 1928 - SOCKS Protocol Version 5.'''

import asyncio
import enum
import logging
import socket
import struct

from slixmpp.stringprep import punycode, StringprepError


log = logging.getLogger(__name__)


class ProtocolMismatch(Exception):
    '''We only implement SOCKS5, no other version or protocol.'''


class ProtocolError(Exception):
    '''Some protocol error.'''


class MethodMismatch(Exception):
    '''The server answered with a method we didn’t ask for.'''


class MethodUnacceptable(Exception):
    '''None of our methods is supported by the server.'''


class AddressTypeUnacceptable(Exception):
    '''The address type (ATYP) field isn’t one of IPv4, IPv6 or domain name.'''


class ReplyError(Exception):
    '''The server answered with an error.'''

    possible_values = (
        "succeeded",
        "general SOCKS server failure",
        "connection not allowed by ruleset",
        "Network unreachable",
        "Host unreachable",
        "Connection refused",
        "TTL expired",
        "Command not supported",
        "Address type not supported",
        "Unknown error")

    def __init__(self, result):
        if result < 9:
            Exception.__init__(self, self.possible_values[result])
        else:
            Exception.__init__(self, self.possible_values[9])


class Method(enum.IntEnum):
    '''Known methods for a SOCKS5 session.'''
    none = 0
    gssapi = 1
    password = 2
    # Methods 3 to 127 are reserved by IANA.
    # Methods 128 to 254 are reserved for private use.
    unacceptable = 255
    not_yet_selected = -1


class Command(enum.IntEnum):
    '''Existing commands for requests.'''
    connect = 1
    bind = 2
    udp_associate = 3


class AddressType(enum.IntEnum):
    '''Existing address types.'''
    ipv4 = 1
    domain = 3
    ipv6 = 4


class Socks5Protocol(asyncio.Protocol):
    '''This implements SOCKS5 as an asyncio protocol.'''

    def __init__(self, dest_addr, dest_port, event):
        self.methods = {Method.none}
        self.selected_method = Method.not_yet_selected
        self.transport = None
        self.dest = (dest_addr, dest_port)
        self.connected = asyncio.Future()
        self.event = event
        self.paused = asyncio.Future()
        self.paused.set_result(None)

    def register_method(self, method):
        '''Register a SOCKS5 method.'''
        self.methods.add(method)

    def unregister_method(self, method):
        '''Unregister a SOCKS5 method.'''
        self.methods.remove(method)

    def connection_made(self, transport):
        '''Called when the connection to the SOCKS5 server is established.'''

        log.debug('SOCKS5 connection established.')

        self.transport = transport
        self._send_methods()

    def data_received(self, data):
        '''Called when we received some data from the SOCKS5 server.'''

        log.debug('SOCKS5 message received.')

        # If we are already connected, this is a data packet.
        if self.connected.done():
            return self.event('socks5_data', data)

        # Every SOCKS5 message starts with the protocol version.
        if data[0] != 5:
            raise ProtocolMismatch()

        # Then select the correct handler for the data we just received.
        if self.selected_method == Method.not_yet_selected:
            self._handle_method(data)
        else:
            self._handle_connect(data)

    def connection_lost(self, exc):
        log.debug('SOCKS5 connection closed.')
        self.event('socks5_closed', exc)

    def pause_writing(self):
        self.paused = asyncio.Future()

    def resume_writing(self):
        self.paused.set_result(None)

    async def write(self, data):
        await self.paused
        self.transport.write(data)

    def _send_methods(self):
        '''Send the methods request, first thing a client should do.'''

        # Create the buffer for our request.
        request = bytearray(len(self.methods) + 2)

        # Protocol version.
        request[0] = 5

        # Number of methods to send.
        request[1] = len(self.methods)

        # List every method we support.
        for i, method in enumerate(self.methods):
            request[i + 2] = method

        # Send the request.
        self.transport.write(request)

    def _send_request(self, command):
        '''Send a request, should be done after having negotiated a method.'''

        # Encode the destination address to embed it in our request.
        # We need to do that first because its length is variable.
        address, port = self.dest
        addr = self._encode_addr(address)

        # Create the buffer for our request.
        request = bytearray(5 + len(addr))

        # Protocol version.
        request[0] = 5

        # Specify the command we want to use.
        request[1] = command

        # request[2] is reserved, keeping it at 0.

        # Add our destination address and port.
        request[3:3+len(addr)] = addr
        request[-2:] = struct.pack('>H', port)

        # Send the request.
        log.debug('SOCKS5 message sent.')
        self.transport.write(request)

    def _handle_method(self, data):
        '''Handle a method reply from the server.'''

        if len(data) != 2:
            raise ProtocolError()
        selected_method = data[1]
        if selected_method not in self.methods:
            raise MethodMismatch()
        if selected_method == Method.unacceptable:
            raise MethodUnacceptable()
        self.selected_method = selected_method
        self._send_request(Command.connect)

    def _handle_connect(self, data):
        '''Handle a connect reply from the server.'''

        try:
            addr, port = self._parse_result(data)
        except ReplyError as exception:
            self.connected.set_exception(exception)
        self.connected.set_result((addr, port))
        self.event('socks5_connected', (addr, port))

    def _parse_result(self, data):
        '''Parse a reply from the server.'''

        result = data[1]
        if result != 0:
            raise ReplyError(result)
        addr = self._parse_addr(data[3:-2])
        port = struct.unpack('>H', data[-2:])[0]
        return (addr, port)

    @staticmethod
    def _parse_addr(addr):
        '''Parse an address (IP or domain) from a bytestream.'''

        addr_type = addr[0]
        if addr_type == AddressType.ipv6:
            try:
                return socket.inet_ntop(socket.AF_INET6, addr[1:])
            except ValueError as e:
                raise AddressTypeUnacceptable(e)
        if addr_type == AddressType.ipv4:
            try:
                return socket.inet_ntop(socket.AF_INET, addr[1:])
            except ValueError as e:
                raise AddressTypeUnacceptable(e)
        if addr_type == AddressType.domain:
            length = addr[1]
            address = addr[2:]
            if length != len(address):
                raise Exception('Size mismatch')
            return address.decode()
        raise AddressTypeUnacceptable(addr_type)

    @staticmethod
    def _encode_addr(addr):
        '''Encode an address (IP or domain) into a bytestream.'''

        try:
            ipv6 = socket.inet_pton(socket.AF_INET6, addr)
            return b'\x04' + ipv6
        except OSError:
            pass
        try:
            ipv4 = socket.inet_aton(addr)
            return b'\x01' + ipv4
        except OSError:
            pass
        try:
            domain = punycode(addr)
            return b'\x03' + bytes([len(domain)]) + domain
        except StringprepError:
            pass
        raise Exception('Err…')



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0020 import stanza
from slixmpp.plugins.xep_0020.stanza import FeatureNegotiation
from slixmpp.plugins.xep_0020.feature_negotiation import XEP_0020


register_plugin(XEP_0020)

__all__ = ['FeatureNegotiation', 'XEP_0020', 'stanza']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.xep_0004.stanza.field import FormField, FieldOption
from slixmpp.plugins.xep_0004.stanza.form import Form




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2015 Emmanuel Gil Peyrot
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class Confirm(ElementBase):

    name = 'confirm'
    namespace = 'http://jabber.org/protocol/http-auth'
    plugin_attrib = 'confirm'
    interfaces = {'id', 'url', 'method'}




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins import BasePlugin, register_plugin


class XEP_0242(BasePlugin):

    name = 'xep_0242'
    description = 'XEP-0242: XMPP Client Compliance 2009'
    dependencies = {'xep_0030', 'xep_0115', 'xep_0054',
                    'xep_0045', 'xep_0085', 'xep_0016',
                    'xep_0191'}


register_plugin(XEP_0242)



# slixmpp.xmlstream.tostring
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# This module converts XML objects into Unicode strings and
# intelligently includes namespaces only when necessary to
# keep the output readable.
# Part of Slixmpp: The Slick XMPP Library
# :copyright: (c) 2011 Nathanael C. Fritz
# :license: MIT, see LICENSE for more details
from __future__ import annotations

from typing import TYPE_CHECKING
from xml.etree.ElementTree import Element
if TYPE_CHECKING:
    from slixmpp.xmlstream import XMLStream

XML_NS = 'http://www.w3.org/XML/1998/namespace'


def tostring(xml: Element | None = None, xmlns: str = '',
             stream: XMLStream | None = None, outbuffer: str = '',
             top_level: bool = False, open_only: bool = False,
             namespaces: set[str] | None = None) -> str:
    """Serialize an XML object to a Unicode string.

    If an outer xmlns is provided using ``xmlns``, then the current element's
    namespace will not be included if it matches the outer namespace. An
    exception is made for elements that have an attached stream, and appear
    at the stream root.

    :param XML xml: The XML object to serialize.
    :param string xmlns: Optional namespace of an element wrapping the XML
                         object.
    :param stream: The XML stream that generated the XML object.
    :param string outbuffer: Optional buffer for storing serializations
                             during recursive calls.
    :param bool top_level: Indicates that the element is the outermost
                           element.
    :param set namespaces: Track which namespaces are in active use so
                           that new ones can be declared when needed.

    :type xml: :py:class:`~xml.etree.ElementTree.Element`
    :type stream: :class:`~slixmpp.xmlstream.xmlstream.XMLStream`

    :rtype: Unicode string
    """
    if xml is None:
        return ''
    # Add previous results to the start of the output.
    output = [outbuffer]

    # Extract the element's tag name.
    tag_split = xml.tag.split('}', 1)
    tag_name = tag_split[-1]

    # Extract the element's namespace if it is defined.
    if '}' in xml.tag:
        tag_xmlns = tag_split[0][1:]
    else:
        tag_xmlns = ''

    default_ns = ''
    stream_ns = ''
    use_cdata = False

    if stream:
        default_ns = stream.default_ns
        stream_ns = stream.stream_ns
        use_cdata = stream.use_cdata

    # Output the tag name and derived namespace of the element.
    namespace = ''
    if tag_xmlns:
        if top_level and tag_xmlns not in [default_ns, xmlns, stream_ns] \
          or not top_level and tag_xmlns != xmlns:
            namespace = ' xmlns="%s"' % tag_xmlns
    if stream and tag_xmlns in stream.namespace_map:
        mapped_namespace = stream.namespace_map[tag_xmlns]
        if mapped_namespace:
            tag_name = "%s:%s" % (mapped_namespace, tag_name)
    output.append("<%s" % tag_name)
    output.append(namespace)

    # Output escaped attribute values.
    new_namespaces = set()
    for attrib, value in xml.attrib.items():
        value = escape(value, use_cdata)
        if '}' not in attrib:
            output.append(' %s="%s"' % (attrib, value))
        else:
            attrib_split = attrib.split('}')
            attrib_ns = attrib_split[0][1:]
            attrib = attrib_split[1]
            if attrib_ns == XML_NS:
                output.append(' xml:%s="%s"' % (attrib, value))
            elif stream and attrib_ns in stream.namespace_map:
                mapped_ns = stream.namespace_map[attrib_ns]
                if mapped_ns:
                    if namespaces is None:
                        namespaces = set()
                    if attrib_ns not in namespaces:
                        namespaces.add(attrib_ns)
                        new_namespaces.add(attrib_ns)
                        output.append(' xmlns:%s="%s"' % (
                            mapped_ns, attrib_ns))
                    output.append(' %s:%s="%s"' % (
                        mapped_ns, attrib, value))

    if open_only:
        # Only output the opening tag, regardless of content.
        output.append(">")
        return ''.join(output)

    if len(xml) or xml.text:
        # If there are additional child elements to serialize.
        output.append(">")
        if xml.text:
            output.append(escape(xml.text, use_cdata))
        if len(xml):
            for child in xml:
                output.append(tostring(child, tag_xmlns, stream,
                    namespaces=namespaces))
        output.append("</%s>" % tag_name)
    elif xml.text:
        # If we only have text content.
        output.append(">%s</%s>" % (escape(xml.text, use_cdata), tag_name))
    else:
        # Empty element.
        output.append(" />")
    if xml.tail:
        # If there is additional text after the element.
        output.append(escape(xml.tail, use_cdata))
    for ns in new_namespaces:
        # Remove namespaces introduced in this context. This is necessary
        # because the namespaces object continues to be shared with other
        # contexts.
        if namespaces is not None:
            namespaces.remove(ns)
    return ''.join(output)


def escape(text: str, use_cdata: bool = False) -> str:
    """Convert special characters in XML to escape sequences.

    :param string text: The XML text to convert.
    :rtype: Unicode string
    """
    escapes = {'&': '&amp;',
               '<': '&lt;',
               '>': '&gt;',
               "'": '&apos;',
               '"': '&quot;'}

    if not use_cdata:
        return ''.join(escapes.get(c, c) for c in text)
    else:
        escape_needed = False
        for c in text:
            if c in escapes:
                escape_needed = True
                break
        if escape_needed:
            escaped = map(lambda x : "<![CDATA[%s]]>" % x, text.split("]]>"))
            return "<![CDATA[]]]><![CDATA[]>]]>".join(escaped)
        return text


def _get_highlight():
    try:
        from pygments import highlight
        from pygments.lexers import get_lexer_by_name
        from pygments.formatters import Terminal256Formatter

        LEXER = get_lexer_by_name('xml')
        FORMATTER = Terminal256Formatter()

        class Highlighter:
            __slots__ = ['string']
            def __init__(self, string):
                self.string = string
            def __str__(self):
                return highlight(str(self.string), LEXER, FORMATTER).strip()

        return Highlighter
    except ImportError:
        return lambda x: x

highlight = _get_highlight()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp import Presence
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0012 import stanza, LastActivity

log = logging.getLogger(__name__)


class XEP_0256(BasePlugin):

    name = 'xep_0256'
    description = 'XEP-0256: Last Activity in Presence'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Presence, LastActivity)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0198.stanza import Enable, Enabled
from slixmpp.plugins.xep_0198.stanza import Resume, Resumed
from slixmpp.plugins.xep_0198.stanza import Failed
from slixmpp.plugins.xep_0198.stanza import StreamManagement
from slixmpp.plugins.xep_0198.stanza import Ack, RequestAck

from slixmpp.plugins.xep_0198.stream_management import XEP_0198


register_plugin(XEP_0198)

__all__ = ['Enable', 'Enabled', 'Resume', 'Resumed', 'Failed',
           'StreamManagement', 'Ack', 'RequestAck', 'XEP_0198']



# Copyright © 2025 Mathieu Pasquet
# This file is part of slixmpp
# See the file LICENSE for copying permission.

from slixmpp.plugins.base import register_plugin
from .sos import XEP_0455, ExternalStatus


register_plugin(XEP_0455)

__all__ = ['XEP_0455', 'ExternalStatus']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.xmlstream import ElementBase


class ReceivedCarbon(ElementBase):
    name = 'received'
    namespace = 'urn:xmpp:carbons:2'
    plugin_attrib = 'carbon_received'
    interfaces = {'carbon_received'}
    is_extension = True

    def get_carbon_received(self):
        return self['forwarded']['stanza']

    def del_carbon_received(self):
        del self['forwarded']['stanza']

    def set_carbon_received(self, stanza):
        self['forwarded']['stanza'] = stanza


class SentCarbon(ElementBase):
    name = 'sent'
    namespace = 'urn:xmpp:carbons:2'
    plugin_attrib = 'carbon_sent'
    interfaces = {'carbon_sent'}
    is_extension = True

    def get_carbon_sent(self):
        return self['forwarded']['stanza']

    def del_carbon_sent(self):
        del self['forwarded']['stanza']

    def set_carbon_sent(self, stanza):
        self['forwarded']['stanza'] = stanza


class PrivateCarbon(ElementBase):
    name = 'private'
    namespace = 'urn:xmpp:carbons:2'
    plugin_attrib = 'carbon_private'
    interfaces = set()


class CarbonEnable(ElementBase):
    name = 'enable'
    namespace = 'urn:xmpp:carbons:2'
    plugin_attrib = 'carbon_enable'
    interfaces = set()


class CarbonDisable(ElementBase):
    name = 'disable'
    namespace = 'urn:xmpp:carbons:2'
    plugin_attrib = 'carbon_disable'
    interfaces = set()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
import logging

from asyncio import Future

from slixmpp import JID
from slixmpp.stanza import Message, Iq
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0280 import stanza


log = logging.getLogger(__name__)


class XEP_0280(BasePlugin):

    """
    XEP-0280 Message Carbons

    Events triggered by this plugin:

    - :term:`carbon_received`
    - :term:`carbon_sent`
    """

    name = 'xep_0280'
    description = 'XEP-0280: Message Carbons'
    dependencies = {'xep_0030', 'xep_0297'}
    stanza = stanza

    def plugin_init(self):
        self.xmpp.register_handler(
            Callback('Carbon Received',
                     StanzaPath('message/carbon_received'),
                     self._handle_carbon_received))
        self.xmpp.register_handler(
            Callback('Carbon Sent',
                     StanzaPath('message/carbon_sent'),
                     self._handle_carbon_sent))

        register_stanza_plugin(Message, stanza.ReceivedCarbon)
        register_stanza_plugin(Message, stanza.SentCarbon)
        register_stanza_plugin(Message, stanza.PrivateCarbon)
        register_stanza_plugin(Iq, stanza.CarbonEnable)
        register_stanza_plugin(Iq, stanza.CarbonDisable)

        register_stanza_plugin(stanza.ReceivedCarbon,
                               self.xmpp['xep_0297'].stanza.Forwarded)
        register_stanza_plugin(stanza.SentCarbon,
                               self.xmpp['xep_0297'].stanza.Forwarded)

    def plugin_end(self):
        self.xmpp.remove_handler('Carbon Received')
        self.xmpp.remove_handler('Carbon Sent')
        self.xmpp.plugin['xep_0030'].del_feature(feature='urn:xmpp:carbons:2')

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature('urn:xmpp:carbons:2')

    def _handle_carbon_received(self, msg: Message):
        if msg['from'].bare == self.xmpp.boundjid.bare:
            self.xmpp.event('carbon_received', msg)

    def _handle_carbon_sent(self, msg: Message):
        if msg['from'].bare == self.xmpp.boundjid.bare:
            self.xmpp.event('carbon_sent', msg)

    def enable(self, ifrom: JID | None = None, **iqkwargs) -> Future:
        """Enable carbons."""
        iq = self.xmpp.make_iq_set(ifrom=ifrom)
        iq.enable('carbon_enable')
        return iq.send(**iqkwargs)

    def disable(self, ifrom: JID | None = None, **iqkwargs) -> Future:
        """Disable carbons."""
        iq = self.xmpp.make_iq_set(ifrom=ifrom)
        iq.enable('carbon_disable')
        return iq.send(**iqkwargs)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Erik Reuterborg Larsson
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging
from asyncio import Future
from typing import Callable

from slixmpp import JID
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0080 import stanza, Geoloc


log = logging.getLogger(__name__)


class XEP_0080(BasePlugin):

    """
    XEP-0080: User Location
    """

    name = 'xep_0080'
    description = 'XEP-0080: User Location'
    dependencies = {'xep_0163'}
    stanza = stanza

    def plugin_end(self):
        self.xmpp['xep_0163'].remove_interest(Geoloc.namespace)
        self.xmpp['xep_0030'].del_feature(feature=Geoloc.namespace)

    def session_bind(self, jid: JID):
        self.xmpp['xep_0163'].register_pep('user_location', Geoloc)

    def publish_location(self, **kwargs) -> Future:
        """
        Publish the user's current location.

        :param accuracy: Horizontal GPS error in meters.
        :param alt: Altitude in meters above or below sea level.
        :param area: A named area such as a campus or neighborhood.
        :param bearing: GPS bearing (direction in which the entity is
                        heading to reach its next waypoint), measured in
                        decimal degrees relative to true north.
        :param building: A specific building on a street or in an area.
        :param country: The nation where the user is located.
        :param countrycode: The ISO 3166 two-letter country code.
        :param datum: GPS datum.
        :param description: A natural-language name for or description of
                            the location.
        :param error: Horizontal GPS error in arc minutes. Obsoleted by
                      the accuracy parameter.
        :param floor: A particular floor in a building.
        :param lat: Latitude in decimal degrees North.
        :param locality: A locality within the administrative region, such
                         as a town or city.
        :param lon: Longitude in decimal degrees East.
        :param postalcode: A code used for postal delivery.
        :param region: An administrative region of the nation, such
                       as a state or province.
        :param room: A particular room in a building.
        :param speed: The speed at which the entity is moving,
                      in meters per second.
        :param street: A thoroughfare within the locality, or a crossing
                       of two thoroughfares.
        :param text: A catch-all element that captures any other
                     information about the location.
        :param timestamp: UTC timestamp specifying the moment when the
                          reading was taken.
        :param uri: A URI or URL pointing to information about
                    the location.
        :param options: Optional form of publish options.
        """
        options = kwargs.get('options', None)
        ifrom = kwargs.get('ifrom', None)
        callback = kwargs.get('callback', None)
        timeout = kwargs.get('timeout', None)
        for param in ('ifrom', 'block', 'callback', 'timeout', 'options'):
            if param in kwargs:
                del kwargs[param]

        geoloc = Geoloc()
        geoloc.values = kwargs

        return self.xmpp['xep_0163'].publish(
            geoloc,
            options=options,
            ifrom=ifrom,
            callback=callback,
            timeout=timeout,
        )

    def stop(self, ifrom: JID | None = None,
             callback: Callable | None = None,
             timeout: int | None = None) -> Future:
        """
        Clear existing user location information to stop notifications.
        """
        geoloc = Geoloc()
        return self.xmpp['xep_0163'].publish(
            geoloc,
            ifrom=ifrom,
            callback=callback,
            timeout=timeout,
        )



from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0065.socks5 import Socks5Protocol
from slixmpp.plugins.xep_0065.stanza import Socks5
from slixmpp.plugins.xep_0065.proxy import XEP_0065


register_plugin(XEP_0065)

__all__ = ['Socks5Protocol', 'Socks5', 'XEP_0065']



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import logging

from asyncio import Future
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0196 import stanza, UserGaming


log = logging.getLogger(__name__)


class XEP_0196(BasePlugin):

    """
    XEP-0196: User Gaming
    """

    name = 'xep_0196'
    description = 'XEP-0196: User Gaming'
    dependencies = {'xep_0163'}
    stanza = stanza

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=UserGaming.namespace)
        self.xmpp['xep_0163'].remove_interest(UserGaming.namespace)

    def session_bind(self, jid):
        self.xmpp['xep_0163'].register_pep('user_gaming', UserGaming)

    def publish_gaming(self, name: str | None = None,
                       level: str | None = None,
                       server_name: str | None = None,
                       uri: str | None = None,
                       character_name: str | None = None,
                       character_profile: str | None = None,
                       server_address: str | None = None,
                       **pubsubkwargs) -> Future:
        """
        Publish the user's current gaming status.

        :param name: The name of the game.
        :param level: The user's level in the game.
        :param uri: A URI for the game or relevant gaming service
        :param server_name: The name of the server where the user is playing.
        :param server_address: The hostname or IP address of the server where the
                               user is playing.
        :param character_name: The name of the user's character in the game.
        :param character_profile: A URI for a profile of the user's character.
        :param options: Optional form of publish options.
        """
        gaming = UserGaming()
        gaming['name'] = name
        gaming['level'] = level
        gaming['uri'] = uri
        gaming['character_name'] = character_name
        gaming['character_profile'] = character_profile
        gaming['server_name'] = server_name
        gaming['server_address'] = server_address
        return self.xmpp['xep_0163'].publish(
            gaming,
            node=UserGaming.namespace,
            **pubsubkwargs
        )

    def stop(self, **pubsubkwargs) -> Future:
        """
        Clear existing user gaming information to stop notifications.
        """
        gaming = UserGaming()
        return self.xmpp['xep_0163'].publish(
            gaming,
            node=UserGaming.namespace,
            **pubsubkwargs
        )




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp import Iq
from slixmpp.xmlstream import register_stanza_plugin, ElementBase, JID
from slixmpp.plugins.xep_0004 import Form
from slixmpp.plugins.xep_0060.stanza.base import OptionalSetting
from slixmpp.plugins.xep_0060.stanza.pubsub import Affiliations, Affiliation
from slixmpp.plugins.xep_0060.stanza.pubsub import Configure, Subscriptions


class PubsubOwner(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#owner'
    name = 'pubsub'
    plugin_attrib = 'pubsub_owner'
    interfaces = set(tuple())


class DefaultConfig(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#owner'
    name = 'default'
    plugin_attrib = name
    interfaces = {'node', 'config'}

    def __init__(self, *args, **kwargs):
        ElementBase.__init__(self, *args, **kwargs)

    def get_config(self):
        return self['form']

    def set_config(self, value):
        del self['from']
        self.append(value)
        return self


class OwnerAffiliations(Affiliations):
    namespace = 'http://jabber.org/protocol/pubsub#owner'
    interfaces = {'node'}

    def append(self, affiliation):
        if not isinstance(affiliation, OwnerAffiliation):
            raise TypeError
        self.xml.append(affiliation.xml)


class OwnerAffiliation(Affiliation):
    namespace = 'http://jabber.org/protocol/pubsub#owner'
    interfaces = {'affiliation', 'jid'}


class OwnerConfigure(Configure):
    namespace = 'http://jabber.org/protocol/pubsub#owner'
    name = 'configure'
    plugin_attrib = name
    interfaces = {'node'}


class OwnerDefault(OwnerConfigure):
    namespace = 'http://jabber.org/protocol/pubsub#owner'
    interfaces = {'node'}


class OwnerDelete(ElementBase, OptionalSetting):
    namespace = 'http://jabber.org/protocol/pubsub#owner'
    name = 'delete'
    plugin_attrib = name
    interfaces = {'node'}


class OwnerPurge(ElementBase, OptionalSetting):
    namespace = 'http://jabber.org/protocol/pubsub#owner'
    name = 'purge'
    plugin_attrib = name
    interfaces = {'node'}


class OwnerRedirect(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#owner'
    name = 'redirect'
    plugin_attrib = name
    interfaces = {'node', 'jid'}

    def set_jid(self, value):
        self._set_attr('jid', str(value))

    def get_jid(self):
        return JID(self._get_attr('jid'))


class OwnerSubscriptions(Subscriptions):
    name = 'subscriptions'
    namespace = 'http://jabber.org/protocol/pubsub#owner'
    plugin_attrib = name
    interfaces = {'node'}

    def append(self, subscription):
        if not isinstance(subscription, OwnerSubscription):
            raise TypeError
        self.xml.append(subscription.xml)


class OwnerSubscription(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub#owner'
    name = 'subscription'
    plugin_attrib = name
    interfaces = {'jid', 'subscription'}

    def set_jid(self, value):
        self._set_attr('jid', str(value))

    def get_jid(self):
        return JID(self._get_attr('jid'))


register_stanza_plugin(Iq, PubsubOwner)
register_stanza_plugin(PubsubOwner, DefaultConfig)
register_stanza_plugin(PubsubOwner, OwnerAffiliations)
register_stanza_plugin(PubsubOwner, OwnerConfigure)
register_stanza_plugin(PubsubOwner, OwnerDefault)
register_stanza_plugin(PubsubOwner, OwnerDelete)
register_stanza_plugin(PubsubOwner, OwnerPurge)
register_stanza_plugin(PubsubOwner, OwnerSubscriptions)
register_stanza_plugin(DefaultConfig, Form)
register_stanza_plugin(OwnerAffiliations, OwnerAffiliation, iterable=True)
register_stanza_plugin(OwnerConfigure, Form)
register_stanza_plugin(OwnerDefault, Form)
register_stanza_plugin(OwnerDelete, OwnerRedirect)
register_stanza_plugin(OwnerSubscriptions, OwnerSubscription, iterable=True)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2021 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0441.stanza import Preferences
from slixmpp.plugins.xep_0441.mam_prefs import XEP_0441


register_plugin(XEP_0441)

__all__ = ['XEP_0441', 'Preferences']



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2025 nicoco
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.plugins.base import register_plugin

from . import stanza
from .notify import XEP_0492

register_plugin(XEP_0492)

__all__ = ["stanza", "XEP_0492"]



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2017 Emmanuel Gil Peyrot
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0300 import stanza
from slixmpp.plugins.xep_0300.stanza import Hash
from slixmpp.plugins.xep_0300.hash import XEP_0300


register_plugin(XEP_0300)

__all__ = ['stanza', 'Hash', 'XEP_0300']




# Slixmpp: The Slick XMPP Library
# Implementation of xeps for Internet of Things
# http://wiki.xmpp.org/web/Tech_pages/IoT_systems
# Copyright (C) 2013 Sustainable Innovation, Joachim.lindborg@sust.se, bjorn.westrom@consoden.se
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging
import time

from functools import partial
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0325 import stanza
from slixmpp.plugins.xep_0325.stanza import Control


log = logging.getLogger(__name__)


class XEP_0325(BasePlugin):

    """
    XEP-0325: IoT Control


    Actuators are devices in sensor networks that can be controlled through
    the network and act with the outside world. In sensor networks and
    Internet of Things applications, actuators make it possible to automate
    real-world processes.
    This plugin implements a mechanism whereby actuators can be controlled
    in XMPP-based sensor networks, making it possible to integrate sensors
    and actuators of different brands, makes and models into larger
    Internet of Things applications.

    Also see <http://xmpp.org/extensions/xep-0325.html>

    Events:
        Sensor side
        -----------
        Control Event:DirectSet    -- Received a control message
        Control Event:SetReq       -- Received a control request

        Client side
        -----------
        Control Event:SetResponse       -- Received a response to a
                                           control request, type result
        Control Event:SetResponseError  -- Received a response to a
                                           control request, type error

    Attributes:
        sessions -- A dictionary or equivalent backend mapping
                    session IDs to dictionaries containing data
                    relevant to a request's session. This dictionary is used
                    both by the client and sensor side. On client side, seqnr
                    is used as key, while on sensor side, a session_id is used
                    as key. This ensures that the two will not collide, so
                    one instance can be both client and sensor.
        Sensor side
        -----------
        nodes    -- A dictionary mapping sensor nodes that are serviced through
                    this XMPP instance to their device handlers ("drivers").
        Client side
        -----------
        last_seqnr -- The last used sequence number (integer). One sequence of
                    communication (e.g. -->request, <--accept, <--fields)
                    between client and sensor is identified by a unique
                    sequence number (unique between the client/sensor pair)

    Methods:
        plugin_init       -- Overrides BasePlugin.plugin_init
        post_init         -- Overrides BasePlugin.post_init
        plugin_end        -- Overrides BasePlugin.plugin_end

        Sensor side
        -----------
        register_node     -- Register a sensor as available from this XMPP
                             instance.

        Client side
        -----------
        set_request       -- Initiates a control request to modify data in
                             sensor(s). Non-blocking, a callback function will
                             be called when the sensor has responded.
        set_command       -- Initiates a control command to modify data in
                             sensor(s). Non-blocking. The sensor(s) will not
                             respond regardless of the result of the command,
                             so no callback is made.

    """

    name = 'xep_0325'
    description = 'XEP-0325 Internet of Things - Control'
    dependencies = {'xep_0030'}
    stanza = stanza


    default_config = {
#        'session_db': None
    }

    def plugin_init(self):
        """ Start the XEP-0325 plugin """

        self.xmpp.register_handler(
                Callback('Control Event:DirectSet',
                    StanzaPath('message/set'),
                    self._handle_direct_set))

        self.xmpp.register_handler(
                Callback('Control Event:SetReq',
                    StanzaPath('iq@type=set/set'),
                    self._handle_set_req))

        self.xmpp.register_handler(
                Callback('Control Event:SetResponse',
                    StanzaPath('iq@type=result/setResponse'),
                    self._handle_set_response))

        self.xmpp.register_handler(
                Callback('Control Event:SetResponseError',
                    StanzaPath('iq@type=error/setResponse'),
                    self._handle_set_response))

        # Server side dicts
        self.nodes = {}
        self.sessions = {}

        self.last_seqnr = 0

        ## For testning only
        self.test_authenticated_from = ""

    def post_init(self):
        """ Init complete. Register our features in Service discovery. """
        BasePlugin.post_init(self)
        self.xmpp['xep_0030'].add_feature(Control.namespace)
        self.xmpp['xep_0030'].set_items(node=Control.namespace, items=tuple())

    def _new_session(self):
        """ Return a new session ID. """
        return str(time.time()) + '-' + self.xmpp.new_id()

    def plugin_end(self):
        """ Stop the XEP-0325 plugin """
        self.sessions.clear()
        self.xmpp.remove_handler('Control Event:DirectSet')
        self.xmpp.remove_handler('Control Event:SetReq')
        self.xmpp.remove_handler('Control Event:SetResponse')
        self.xmpp.remove_handler('Control Event:SetResponseError')
        self.xmpp['xep_0030'].del_feature(feature=Control.namespace)
        self.xmpp['xep_0030'].set_items(node=Control.namespace, items=tuple())


    # =================================================================
    # Sensor side (data provider) API

    def register_node(self, nodeId, device, commTimeout, sourceId=None, cacheType=None):
        """
        Register a sensor/device as available for control requests/commands
        through this XMPP instance.

        The device object may by any custom implementation to support
        specific devices, but it must implement the functions:
          has_control_field
          set_control_fields
        according to the interfaces shown in the example device.py file.

        Arguments:
            nodeId      -- The identifier for the device
            device      -- The device object
            commTimeout -- Time in seconds to wait between each callback from device during
                           a data readout. Float.
            sourceId    -- [optional] identifying the data source controlling the device
            cacheType   -- [optional] narrowing down the search to a specific kind of node
        """
        self.nodes[nodeId] = {"device": device,
                                "commTimeout": commTimeout,
                                "sourceId": sourceId,
                                "cacheType": cacheType}

    def _set_authenticated(self, auth=''):
        """ Internal testing function """
        self.test_authenticated_from = auth

    def _get_new_seqnr(self):
        """ Returns a unique sequence number (unique across threads) """
        self.last_seqnr = self.last_seqnr + 1
        return str(self.last_seqnr)

    def _handle_set_req(self, iq):
        """
        Event handler for reception of an Iq with set req - this is a
        control request.

        Verifies that
          - all the requested nodes are available
            (if no nodes are specified in the request, assume all nodes)
          - all the control fields are available from all requested nodes
            (if no nodes are specified in the request, assume all nodes)

        If the request passes verification, the control request is passed
        to the devices (in a separate thread).
        If the verification fails, a setResponse with error indication
        is sent.
        """

        error_msg = ''
        req_ok = True
        missing_node = None
        missing_field = None

        # Authentication
        if len(self.test_authenticated_from) > 0 and not iq['from'] == self.test_authenticated_from:
            # Invalid authentication
            req_ok = False
            error_msg = "Access denied"

        # Nodes
        if len(iq['set']['nodes']) > 0:
            for n in iq['set']['nodes']:
                if n['nodeId'] not in self.nodes:
                    req_ok = False
                    missing_node = n['nodeId']
                    error_msg = "Invalid nodeId " + n['nodeId']
            process_nodes = [n['nodeId'] for n in iq['set']['nodes']]
        else:
            process_nodes = self.nodes.keys()

        # Fields - for control we need to find all in all devices, otherwise we reject
        process_fields = []
        if len(iq['set']['datas']) > 0:
            for f in iq['set']['datas']:
                for node in self.nodes:
                    if not self.nodes[node]["device"].has_control_field(f['name'], f._get_typename()):
                        req_ok = False
                        missing_field = f['name']
                        error_msg = "Invalid field " + f['name']
                        break
            process_fields = [(f['name'], f._get_typename(), f['value']) for f in iq['set']['datas']]

        if req_ok:
            session = self._new_session()
            self.sessions[session] = {"from": iq['from'], "to": iq['to'], "seqnr": iq['id']}
            self.sessions[session]["commTimers"] = {}
            self.sessions[session]["nodeDone"] = {}
            # Flag that a reply is executed when we are done
            self.sessions[session]["reply"] = True

            self.sessions[session]["node_list"] = process_nodes
            self._node_request(session, process_fields)
        else:
            iq = iq.reply()
            iq['type'] = 'error'
            iq['setResponse']['responseCode'] = "NotFound"
            if missing_node is not None:
                iq['setResponse'].add_node(missing_node)
            if missing_field is not None:
                iq['setResponse'].add_data(missing_field)
            iq['setResponse']['error']['var'] = "Output"
            iq['setResponse']['error']['text'] = error_msg
            iq.send()

    def _handle_direct_set(self, msg):
        """
        Event handler for reception of a Message with set command - this is a
        direct control command.

        Verifies that
          - all the requested nodes are available
            (if no nodes are specified in the request, assume all nodes)
          - all the control fields are available from all requested nodes
            (if no nodes are specified in the request, assume all nodes)

        If the request passes verification, the control request is passed
        to the devices (in a separate thread).
        If the verification fails, do nothing.
        """
        req_ok = True

        # Nodes
        if len(msg['set']['nodes']) > 0:
            for n in msg['set']['nodes']:
                if n['nodeId'] not in self.nodes:
                    req_ok = False
                    error_msg = "Invalid nodeId " + n['nodeId']
            process_nodes = [n['nodeId'] for n in msg['set']['nodes']]
        else:
            process_nodes = self.nodes.keys()

        # Fields - for control we need to find all in all devices, otherwise we reject
        process_fields = []
        if len(msg['set']['datas']) > 0:
            for f in msg['set']['datas']:
                for node in self.nodes:
                    if not self.nodes[node]["device"].has_control_field(f['name'], f._get_typename()):
                        req_ok = False
                        missing_field = f['name']
                        error_msg = "Invalid field " + f['name']
                        break
            process_fields = [(f['name'], f._get_typename(), f['value']) for f in msg['set']['datas']]

        if req_ok:
            session = self._new_session()
            self.sessions[session] = {"from": msg['from'], "to": msg['to']}
            self.sessions[session]["commTimers"] = {}
            self.sessions[session]["nodeDone"] = {}
            self.sessions[session]["reply"] = False

            self.sessions[session]["node_list"] = process_nodes
            self._node_request(session, process_fields)


    def _node_request(self, session, process_fields):
        """
        Helper function to handle the device control in a separate thread.

        Arguments:
            session         -- The request session id
            process_fields  -- The fields to set in the devices. List of tuple format:
                               (name, datatype, value)
        """
        for node in self.sessions[session]["node_list"]:
            self.sessions[session]["nodeDone"][node] = False

        for node in self.sessions[session]["node_list"]:
            timer = self.xmpp.loop.call_later(self.nodes[node]['commTimeout'], partial(self._event_comm_timeout, args=(session, node)))
            self.sessions[session]["commTimers"][node] = timer
            self.nodes[node]['device'].set_control_fields(process_fields, session=session, callback=self._device_set_command_callback)

    def _event_comm_timeout(self, session, nodeId):
        """
        Triggered if any of the control operations timeout.
        Stop communicating with the failing device.
        If the control command was an Iq request, sends a failure
        message back to the client.

        Arguments:
            session         -- The request session id
            nodeId          -- The id of the device which timed out
        """

        if self.sessions[session]["reply"]:
            # Reply is executed when we are done
            iq = self.xmpp.Iq()
            iq['from'] = self.sessions[session]['to']
            iq['to'] = self.sessions[session]['from']
            iq['type'] = "error"
            iq['id'] = self.sessions[session]['seqnr']
            iq['setResponse']['responseCode'] = "OtherError"
            iq['setResponse'].add_node(nodeId)
            iq['setResponse']['error']['var'] = "Output"
            iq['setResponse']['error']['text'] = "Timeout."
            iq.send()

        ## TODO - should we send one timeout per node??

        # Drop communication with this device and check if we are done
        self.sessions[session]["nodeDone"][nodeId] = True
        if (self._all_nodes_done(session)):
            # The session is complete, delete it
            del self.sessions[session]

    def _all_nodes_done(self, session):
        """
        Checks whether all devices are done replying to the control command.

        Arguments:
            session         -- The request session id
        """
        for n in self.sessions[session]["nodeDone"]:
            if not self.sessions[session]["nodeDone"][n]:
                return False
        return True

    def _device_set_command_callback(self, session, nodeId, result, error_field=None, error_msg=None):
        """
        Callback function called by the devices when the control command is
        complete or failed.
        If needed, composes a message with the result and sends it back to the
        client.

        Arguments:
            session         -- The request session id
            nodeId          -- The device id which initiated the callback
            result          -- The current result status of the control command. Valid values are:
                               "error"  - Set fields failed.
                               "ok"     - All fields were set.
            error_field      -- [optional] Only applies when result == "error"
                                The field name that failed (usually means it is missing)
            error_msg        -- [optional] Only applies when result == "error".
                                Error details when a request failed.
        """

        if session not in self.sessions:
            # This can happen if a session was deleted, like in a timeout. Just drop the data.
            return

        if result == "error":
            self.sessions[session]["commTimers"][nodeId].cancel()

            if self.sessions[session]["reply"]:
                # Reply is executed when we are done
                iq = self.xmpp.Iq()
                iq['from'] = self.sessions[session]['to']
                iq['to'] = self.sessions[session]['from']
                iq['type'] = "error"
                iq['id'] = self.sessions[session]['seqnr']
                iq['setResponse']['responseCode'] = "OtherError"
                iq['setResponse'].add_node(nodeId)
                if error_field is not None:
                    iq['setResponse'].add_data(error_field)
                iq['setResponse']['error']['var'] = error_field
                iq['setResponse']['error']['text'] = error_msg
                iq.send()

                # Drop communication with this device and check if we are done
                self.sessions[session]["nodeDone"][nodeId] = True
                if (self._all_nodes_done(session)):
                    # The session is complete, delete it
                    del self.sessions[session]
        else:
            self.sessions[session]["commTimers"][nodeId].cancel()

            self.sessions[session]["nodeDone"][nodeId] = True
            if (self._all_nodes_done(session)):
                if self.sessions[session]["reply"]:
                    # Reply is executed when we are done
                    iq = self.xmpp.Iq()
                    iq['from'] = self.sessions[session]['to']
                    iq['to'] = self.sessions[session]['from']
                    iq['type'] = "result"
                    iq['id'] = self.sessions[session]['seqnr']
                    iq['setResponse']['responseCode'] = "OK"
                    iq.send()

                # The session is complete, delete it
                del self.sessions[session]


    # =================================================================
    # Client side (data controller) API

    def set_request(self, from_jid, to_jid, callback, fields, nodeIds=None):
        """
        Called on the client side to initiade a control request.
        Composes a message with the request and sends it to the device(s).
        Does not block, the callback will be called when the device(s)
        has responded.

        Arguments:
            from_jid        -- The jid of the requester
            to_jid          -- The jid of the device(s)
            callback        -- The callback function to call when data is available.

                            The callback function must support the following arguments:

                from_jid    -- The jid of the responding device(s)
                result      -- The result of the control request. Valid values are:
                               "OK"             - Control request completed successfully
                               "NotFound"       - One or more nodes or fields are missing
                               "InsufficientPrivileges" - Not authorized.
                               "Locked"         - Field(s) is locked and cannot
                                                  be changed at the moment.
                               "NotImplemented" - Request feature not implemented.
                               "FormError"      - Error while setting with
                                                  a form (not implemented).
                               "OtherError"     - Indicates other types of
                                                  errors, such as timeout.
                                                  Details in the error_msg.


                nodeId      -- [optional] Only applicable when result == "error"
                               List of node Ids of failing device(s).

                fields      -- [optional] Only applicable when result == "error"
                               List of fields that failed.[optional] Mandatory when result == "rejected" or "failure".

                error_msg   -- Details about why the request failed.

            fields          -- Fields to set. List of tuple format: (name, typename, value).
            nodeIds         -- [optional] Limits the request to the node Ids in this list.
        """
        iq = self.xmpp.Iq()
        iq['from'] = from_jid
        iq['to'] = to_jid
        seqnr = self._get_new_seqnr()
        iq['id'] = seqnr
        iq['type'] = "set"
        if nodeIds is not None:
            for nodeId in nodeIds:
                iq['set'].add_node(nodeId)
        if fields is not None:
            for name, typename, value in fields:
                iq['set'].add_data(name=name, typename=typename, value=value)

        self.sessions[seqnr] = {"from": iq['from'], "to": iq['to'], "callback": callback}
        iq.send()

    def set_command(self, from_jid, to_jid, fields, nodeIds=None):
        """
        Called on the client side to initiade a control command.
        Composes a message with the set commandand sends it to the device(s).
        Does not block. Device(s) will not respond, regardless of result.

        Arguments:
            from_jid        -- The jid of the requester
            to_jid          -- The jid of the device(s)

            fields          -- Fields to set. List of tuple format: (name, typename, value).
            nodeIds         -- [optional] Limits the request to the node Ids in this list.
        """
        msg = self.xmpp.Message()
        msg['from'] = from_jid
        msg['to'] = to_jid
        msg['type'] = "set"
        if nodeIds is not None:
            for nodeId in nodeIds:
                msg['set'].add_node(nodeId)
        if fields is not None:
            for name, typename, value in fields:
                msg['set'].add_data(name, typename, value)

        # We won't get any reply, so don't create a session
        msg.send()

    def _handle_set_response(self, iq):
        """ Received response from device(s) """
        #print("ooh")
        seqnr = iq['id']
        from_jid = str(iq['from'])
        result = iq['setResponse']['responseCode']
        nodeIds = [n['name'] for n in iq['setResponse']['nodes']]
        fields = [f['name'] for f in iq['setResponse']['datas']]
        error_msg = None

        if iq['setResponse'].xml.find('error') is not None and not iq['setResponse']['error']['text'] == "":
            error_msg = iq['setResponse']['error']['text']

        callback = self.sessions[seqnr]["callback"]
        callback(from_jid=from_jid, result=result, nodeIds=nodeIds, fields=fields, error_msg=error_msg)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0422 import stanza


class XEP_0422(BasePlugin):
    '''XEP-0422: Message Fastening'''

    name = 'xep_0422'
    description = 'XEP-0422: Message Fastening'
    dependencies = {'xep_0030'}
    stanza = stanza
    namespace = stanza.NS

    def plugin_init(self) -> None:
        stanza.register_plugins()

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature(feature=stanza.NS)

    def plugin_end(self):
        self.xmpp.plugin['xep_0030'].del_feature(feature=stanza.NS)



"""
Usage:

>>> from slixmpp import Iq
>>> register_stanza_plugin(Iq, DiscoItems)  # automatically done when loading
>>> register_plugin()                       # the XEP_0462 plugin
>>> iq = Iq()
>>> iq["disco_items"]["filter"]["included_types"] = ["type1", "type2"]
>>> iq.pretty_print()
<iq xmlns="jabber:client" id="0">
  <query xmlns="http://jabber.org/protocol/disco#items">
    <filter xmlns="urn:xmpp:pubsub-filter:0">
      <x xmlns="jabber:x:data" type="submit">
        <field var="FORM_TYPE" type="hidden">
          <value>urn:xmpp:pubsub-filter:0</value>
        </field>
        <field var="included-types">
          <value>type1</value>
          <value>type2</value>
        </field>
      </x>
    </filter>
  </query>
</iq>
>>> iq["disco_items"]["filter"]["included_types"]
['type1', 'type2']

And similarly, an "excluded_types" interface is present.
"""

from slixmpp.plugins.xep_0004 import Form
from slixmpp.plugins.xep_0030.stanza.items import DiscoItems
from slixmpp.xmlstream import ElementBase, register_stanza_plugin

NS = "urn:xmpp:pubsub-filter:0"


class Filter(ElementBase):
    namespace = NS
    name = "filter"
    plugin_attrib = "filter"
    interfaces = {"included_types", "excluded_types"}

    def set_included_types(self, types: list[str]) -> None:
        self.__set("in", types)

    def get_included_types(self) -> list[str]:
        return self.__get("in")

    def set_excluded_types(self, types: list[str]) -> None:
        self.__set("ex", types)

    def get_excluded_types(self) -> list[str]:
        return self.__get("ex")

    def __get(self, what: str) -> list[str]:
        try:
            return self["form"].field[f"{what}cluded-types"].values["value"]
        except KeyError:
            return []

    def __set(self, what: str, types: list[str]) -> None:
        if self.get_plugin("form", check=True):
            form: Form = self["form"]
            if f"{what}cluded-types" in form.field:
                form.field[f"{what}cluded-types"]["value"] = types
            else:
                form.add_field(
                    var="{what}cluded-types",
                    ftype="list-multi",
                    value=types,
                )
        else:
            form: Form = self["form"]
            form["type"] = "submit"
            form.add_field(var="FORM_TYPE", ftype="hidden", value=NS)
            form.add_field(
                var=f"{what}cluded-types",
                ftype="list-multi",
                value=types,
            )


def register_plugin():
    register_stanza_plugin(DiscoItems, Filter)
    register_stanza_plugin(Filter, Form)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class LastActivity(ElementBase):

    name = 'query'
    namespace = 'jabber:iq:last'
    plugin_attrib = 'last_activity'
    interfaces = {'seconds', 'status'}

    def get_seconds(self):
        return int(self._get_attr('seconds'))

    def set_seconds(self, value):
        self._set_attr('seconds', str(value))

    def get_status(self):
        return self.xml.text

    def set_status(self, value):
        self.xml.text = str(value)

    def del_status(self):
        self.xml.text = ''




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream.stanzabase import StanzaBase
from slixmpp.xmlstream.handler import Waiter


class XMLWaiter(Waiter):

    """
    The XMLWaiter class is identical to the normal Waiter class
    except that it returns the XML contents of the stanza instead
    of the full stanza object itself.

    Methods:
        prerun -- Overrides Waiter.prerun
    """

    def prerun(self, payload: StanzaBase) -> None:
        """
        Store the XML contents of the stanza to return to the
        waiting event handler.

        Overrides Waiter.prerun

        Arguments:
            payload -- The matched stanza object.
        """
        Waiter.prerun(self, payload.xml)  # type: ignore




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import StanzaBase, ET


class Failure(StanzaBase):

    """
    """

    name = 'failure'
    namespace = 'urn:ietf:params:xml:ns:xmpp-sasl'
    interfaces = {'condition', 'text'}
    plugin_attrib = name
    sub_interfaces = {'text'}
    conditions = {'aborted', 'account-disabled', 'credentials-expired',
                  'encryption-required', 'incorrect-encoding',
                  'invalid-authzid', 'invalid-mechanism', 'malformed-request',
                  'mechansism-too-weak', 'not-authorized',
                  'temporary-auth-failure'}

    def setup(self, xml=None):
        """
        Populate the stanza object using an optional XML object.

        Overrides ElementBase.setup.

        Sets a default error type and condition, and changes the
        parent stanza's type to 'error'.

        Arguments:
            xml -- Use an existing XML object for the stanza's values.
        """
        # StanzaBase overrides self.namespace
        self.namespace = Failure.namespace

        if StanzaBase.setup(self, xml):
            #If we had to generate XML then set default values.
            self['condition'] = 'not-authorized'

        self.xml.tag = self.tag_name()

    def get_condition(self):
        """Return the condition element's name."""
        for child in self.xml:
            if "{%s}" % self.namespace in child.tag:
                cond = child.tag.split('}', 1)[-1]
                if cond in self.conditions:
                    return cond
        return 'not-authorized'

    def set_condition(self, value):
        """
        Set the tag name of the condition element.

        Arguments:
           value -- The tag name of the condition element.
        """
        if value in self.conditions:
            del self['condition']
            self.xml.append(ET.Element("{%s}%s" % (self.namespace, value)))
        return self

    def del_condition(self):
        """Remove the condition element."""
        for child in self.xml:
            if "{%s}" % self.namespace in child.tag:
                tag = child.tag.split('}', 1)[-1]
                if tag in self.conditions:
                    self.xml.remove(child)
        return self



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0152 import stanza
from slixmpp.plugins.xep_0152.stanza import Reachability
from slixmpp.plugins.xep_0152.reachability import XEP_0152


register_plugin(XEP_0152)

__all__ = ['stanza', 'Reachability', 'XEP_0152']



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0115.stanza import Capabilities
from slixmpp.plugins.xep_0115.static import StaticCaps
from slixmpp.plugins.xep_0115.caps import XEP_0115


register_plugin(XEP_0115)

__all__ = ['Capabilities', 'StaticCaps', 'XEP_0115']




# Slixmpp: The Slick XMPP Library
# Implementation of xeps for Internet of Things
# http://wiki.xmpp.org/web/Tech_pages/IoT_systems
# Copyright (C) 2013 Sustainable Innovation, Joachim.lindborg@sust.se, bjorn.westrom@consoden.se
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging
import time
import datetime
from threading import Thread, Lock, Timer

from slixmpp.plugins.xep_0323.timerreset import TimerReset

from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0323 import stanza
from slixmpp.plugins.xep_0323.stanza import Sensordata

log = logging.getLogger(__name__)


class XEP_0323(BasePlugin):

    """
    XEP-0323: IoT Sensor Data


    This XEP provides the underlying architecture, basic operations and data
    structures for sensor data communication over XMPP networks. It includes
    a hardware abstraction model, removing any technical detail implemented
    in underlying technologies.

    Also see <http://xmpp.org/extensions/xep-0323.html>

    Configuration Values:
        threaded -- Indicates if communication with sensors should be threaded.
                    Defaults to True.

    Events:
        Sensor side
        -----------
        Sensordata Event:Req    -- Received a request for data
        Sensordata Event:Cancel -- Received a cancellation for a request

        Client side
        -----------
        Sensordata Event:Accepted -- Received a accept from sensor for a request
        Sensordata Event:Rejected -- Received a reject from sensor for a request
        Sensordata Event:Cancelled -- Received a cancel confirm from sensor
        Sensordata Event:Fields   -- Received fields from sensor for a request
                                     This may be triggered multiple times since
                                     the sensor can split up its response in
                                     multiple messages.
        Sensordata Event:Failure  -- Received a failure indication from sensor
                                     for a request. Typically a comm timeout.

    Attributes:
        threaded -- Indicates if command events should be threaded.
                    Defaults to True.
        sessions -- A dictionary or equivalent backend mapping
                    session IDs to dictionaries containing data
                    relevant to a request's session. This dictionary is used
                    both by the client and sensor side. On client side, seqnr
                    is used as key, while on sensor side, a session_id is used
                    as key. This ensures that the two will not collide, so
                    one instance can be both client and sensor.
        Sensor side
        -----------
        nodes    -- A dictionary mapping sensor nodes that are serviced through
                    this XMPP instance to their device handlers ("drivers").
        Client side
        -----------
        last_seqnr -- The last used sequence number (integer). One sequence of
                    communication (e.g. -->request, <--accept, <--fields)
                    between client and sensor is identified by a unique
                    sequence number (unique between the client/sensor pair)

    Methods:
        plugin_init       -- Overrides BasePlugin.plugin_init
        post_init         -- Overrides BasePlugin.post_init
        plugin_end        -- Overrides BasePlugin.plugin_end

        Sensor side
        -----------
        register_node     -- Register a sensor as available from this XMPP
                             instance.

        Client side
        -----------
        request_data      -- Initiates a request for data from one or more
                             sensors. Non-blocking, a callback function will
                             be called when data is available.

    """

    name = 'xep_0323'
    description = 'XEP-0323 Internet of Things - Sensor Data'
    dependencies = {'xep_0030'}
    stanza = stanza


    default_config = {
        'threaded': True
    }

    def plugin_init(self):
        """ Start the XEP-0323 plugin """

        self.xmpp.register_handler(
                Callback('Sensordata Event:Req',
                    StanzaPath('iq@type=get/req'),
                    self._handle_event_req))

        self.xmpp.register_handler(
                Callback('Sensordata Event:Accepted',
                    StanzaPath('iq@type=result/accepted'),
                    self._handle_event_accepted))

        self.xmpp.register_handler(
                Callback('Sensordata Event:Rejected',
                    StanzaPath('iq@type=error/rejected'),
                    self._handle_event_rejected))

        self.xmpp.register_handler(
                Callback('Sensordata Event:Cancel',
                    StanzaPath('iq@type=get/cancel'),
                    self._handle_event_cancel))

        self.xmpp.register_handler(
                Callback('Sensordata Event:Cancelled',
                    StanzaPath('iq@type=result/cancelled'),
                    self._handle_event_cancelled))

        self.xmpp.register_handler(
                Callback('Sensordata Event:Fields',
                    StanzaPath('message/fields'),
                    self._handle_event_fields))

        self.xmpp.register_handler(
                Callback('Sensordata Event:Failure',
                    StanzaPath('message/failure'),
                    self._handle_event_failure))

        self.xmpp.register_handler(
                Callback('Sensordata Event:Started',
                    StanzaPath('message/started'),
                    self._handle_event_started))

        # Server side dicts
        self.nodes = {}
        self.sessions = {}

        self.last_seqnr = 0
        self.seqnr_lock = Lock()

        ## For testing only
        self.test_authenticated_from = ""

    def post_init(self):
        """ Init complete. Register our features in Service discovery. """
        BasePlugin.post_init(self)
        self.xmpp['xep_0030'].add_feature(Sensordata.namespace)
        self.xmpp['xep_0030'].set_items(node=Sensordata.namespace, items=tuple())

    def _new_session(self):
        """ Return a new session ID. """
        return str(time.time()) + '-' + self.xmpp.new_id()

    def session_bind(self, jid):
        logging.debug("setting the Disco discovery for %s" % Sensordata.namespace)
        self.xmpp['xep_0030'].add_feature(Sensordata.namespace)
        self.xmpp['xep_0030'].set_items(node=Sensordata.namespace, items=tuple())


    def plugin_end(self):
        """ Stop the XEP-0323 plugin """
        self.sessions.clear()
        self.xmpp.remove_handler('Sensordata Event:Req')
        self.xmpp.remove_handler('Sensordata Event:Accepted')
        self.xmpp.remove_handler('Sensordata Event:Rejected')
        self.xmpp.remove_handler('Sensordata Event:Cancel')
        self.xmpp.remove_handler('Sensordata Event:Cancelled')
        self.xmpp.remove_handler('Sensordata Event:Fields')
        self.xmpp['xep_0030'].del_feature(feature=Sensordata.namespace)


    # =================================================================
    # Sensor side (data provider) API

    def register_node(self, nodeId, device, commTimeout, sourceId=None, cacheType=None):
        """
        Register a sensor/device as available for serving of data through this XMPP
        instance.

        The device object may by any custom implementation to support
        specific devices, but it must implement the functions:
          has_field
          request_fields
        according to the interfaces shown in the example device.py file.

        Arguments:
            nodeId      -- The identifier for the device
            device      -- The device object
            commTimeout -- Time in seconds to wait between each callback from device during
                           a data readout. Float.
            sourceId    -- [optional] identifying the data source controlling the device
            cacheType   -- [optional] narrowing down the search to a specific kind of node
        """
        self.nodes[nodeId] = {"device": device,
                                "commTimeout": commTimeout,
                                "sourceId": sourceId,
                                "cacheType": cacheType}

    def _set_authenticated(self, auth=''):
        """ Internal testing function """
        self.test_authenticated_from = auth


    def _handle_event_req(self, iq):
        """
        Event handler for reception of an Iq with req - this is a request.

        Verifies that
          - all the requested nodes are available
          - at least one of the requested fields is available from at least
            one of the nodes

        If the request passes verification, an accept response is sent, and
        the readout process is started in a separate thread.
        If the verification fails, a reject message is sent.
        """

        seqnr = iq['req']['seqnr']
        error_msg = ''
        req_ok = True

        # Authentication
        if len(self.test_authenticated_from) > 0 and not iq['from'] == self.test_authenticated_from:
            # Invalid authentication
            req_ok = False
            error_msg = "Access denied"

        # Nodes
        process_nodes = []
        if len(iq['req']['nodes']) > 0:
            for n in iq['req']['nodes']:
                if n['nodeId'] not in self.nodes:
                    req_ok = False
                    error_msg = "Invalid nodeId " + n['nodeId']
            process_nodes = [n['nodeId'] for n in iq['req']['nodes']]
        else:
            process_nodes = self.nodes.keys()

        # Fields - if we just find one we are happy, otherwise we reject
        process_fields = []
        if len(iq['req']['fields']) > 0:
            found = False
            for f in iq['req']['fields']:
                for node in self.nodes:
                    if self.nodes[node]["device"].has_field(f['name']):
                        found = True
                        break
            if not found:
                req_ok = False
                error_msg = "Invalid field " + f['name']
            process_fields = [f['name'] for n in iq['req']['fields']]

        req_flags = iq['req']._get_flags()

        request_delay_sec = None
        if 'when' in req_flags:
            # Timed request - requires datetime string in iso format
            # ex. 2013-04-05T15:00:03
            dt = None
            try:
                dt = datetime.datetime.strptime(req_flags['when'], "%Y-%m-%dT%H:%M:%S")
            except ValueError:
                req_ok = False
                error_msg = "Invalid datetime in 'when' flag, please use ISO format (i.e. 2013-04-05T15:00:03)."

            if dt is not None:
                # Datetime properly formatted
                dtnow = datetime.datetime.now()
                dtdiff = dt - dtnow
                request_delay_sec = dtdiff.seconds + dtdiff.days * 24 * 3600
                if request_delay_sec <= 0:
                    req_ok = False
                    error_msg = "Invalid datetime in 'when' flag, cannot set a time in the past (%s). Current time: %s" % (dt.isoformat(), dtnow.isoformat())

        if req_ok:
            session = self._new_session()
            self.sessions[session] = {"from": iq['from'], "to": iq['to'], "seqnr": seqnr}
            self.sessions[session]["commTimers"] = {}
            self.sessions[session]["nodeDone"] = {}

            iq = iq.reply()
            iq['accepted']['seqnr'] = seqnr
            if request_delay_sec is not None:
                iq['accepted']['queued'] = "true"
            iq.send()

            self.sessions[session]["node_list"] = process_nodes

            if request_delay_sec is not None:
                # Delay request to requested time
                timer = Timer(request_delay_sec, self._event_delayed_req, args=(session, process_fields, req_flags))
                self.sessions[session]["commTimers"]["delaytimer"] = timer
                timer.start()
                return

            if self.threaded:
                tr_req = Thread(target=self._threaded_node_request, args=(session, process_fields, req_flags))
                tr_req.start()
            else:
                self._threaded_node_request(session, process_fields, req_flags)

        else:
            iq = iq.reply()
            iq['type'] = 'error'
            iq['rejected']['seqnr'] = seqnr
            iq['rejected']['error'] = error_msg
            iq.send()

    def _threaded_node_request(self, session, process_fields, flags):
        """
        Helper function to handle the device readouts in a separate thread.

        Arguments:
            session         -- The request session id
            process_fields  -- The fields to request from the devices
            flags           -- [optional] flags to pass to the devices, e.g. momentary
                               Formatted as a dictionary like { "flag name": "flag value" ... }
        """
        for node in self.sessions[session]["node_list"]:
            self.sessions[session]["nodeDone"][node] = False

        for node in self.sessions[session]["node_list"]:
            timer = TimerReset(self.nodes[node]['commTimeout'], self._event_comm_timeout, args=(session, node))
            self.sessions[session]["commTimers"][node] = timer
            timer.start()
            self.nodes[node]['device'].request_fields(process_fields, flags=flags, session=session, callback=self._device_field_request_callback)

    def _event_comm_timeout(self, session, nodeId):
        """
        Triggered if any of the readout operations timeout.
        Sends a failure message back to the client, stops communicating
        with the failing device.

        Arguments:
            session         -- The request session id
            nodeId          -- The id of the device which timed out
        """
        msg = self.xmpp.Message()
        msg['from'] = self.sessions[session]['to']
        msg['to'] = self.sessions[session]['from']
        msg['failure']['seqnr'] = self.sessions[session]['seqnr']
        msg['failure']['error']['text'] = "Timeout"
        msg['failure']['error']['nodeId'] = nodeId
        msg['failure']['error']['timestamp'] = datetime.datetime.now().replace(microsecond=0).isoformat()

        # Drop communication with this device and check if we are done
        self.sessions[session]["nodeDone"][nodeId] = True
        if (self._all_nodes_done(session)):
            msg['failure']['done'] = 'true'
        msg.send()
        # The session is complete, delete it
        del self.sessions[session]

    def _event_delayed_req(self, session, process_fields, req_flags):
        """
        Triggered when the timer from a delayed request fires.

        Arguments:
            session         -- The request session id
            process_fields  -- The fields to request from the devices
            flags           -- [optional] flags to pass to the devices, e.g. momentary
                               Formatted as a dictionary like { "flag name": "flag value" ... }
        """
        msg = self.xmpp.Message()
        msg['from'] = self.sessions[session]['to']
        msg['to'] = self.sessions[session]['from']
        msg['started']['seqnr'] = self.sessions[session]['seqnr']
        msg.send()

        if self.threaded:
            tr_req = Thread(target=self._threaded_node_request, args=(session, process_fields, req_flags))
            tr_req.start()
        else:
            self._threaded_node_request(session, process_fields, req_flags)

    def _all_nodes_done(self, session):
        """
        Checks whether all devices are done replying to the readout.

        Arguments:
            session         -- The request session id
        """
        for n in self.sessions[session]["nodeDone"]:
            if not self.sessions[session]["nodeDone"][n]:
                return False
        return True

    def _device_field_request_callback(self, session, nodeId, result, timestamp_block, error_msg=None):
        """
        Callback function called by the devices when they have any additional data.
        Composes a message with the data and sends it back to the client, and resets
        the timeout timer for the device.

        Arguments:
            session         -- The request session id
            nodeId          -- The device id which initiated the callback
            result          -- The current result status of the readout. Valid values are:
                               "error"  - Readout failed.
                               "fields" - Contains readout data.
                               "done"   - Indicates that the readout is complete. May contain
                                          readout data.
            timestamp_block -- [optional] Only applies when result != "error"
                               The readout data. Structured as a dictionary:
              {
                timestamp:     timestamp for this datablock,
                fields:        list of field dictionary (one per readout field).
                  readout field dictionary format:
                  {
                    type:      The field type (numeric, boolean, dateTime, timeSpan, string, enum)
                    name:      The field name
                    value:     The field value
                    unit:      The unit of the field. Only applies to type numeric.
                    dataType:  The datatype of the field. Only applies to type enum.
                    flags:     [optional] data classifier flags for the field, e.g. momentary
                               Formatted as a dictionary like { "flag name": "flag value" ... }
                  }
              }
            error_msg        -- [optional] Only applies when result == "error".
                                Error details when a request failed.
        """
        if session not in self.sessions:
            # This can happen if a session was deleted, like in a cancellation. Just drop the data.
            return

        if result == "error":
            self.sessions[session]["commTimers"][nodeId].cancel()

            msg = self.xmpp.Message()
            msg['from'] = self.sessions[session]['to']
            msg['to'] = self.sessions[session]['from']
            msg['failure']['seqnr'] = self.sessions[session]['seqnr']
            msg['failure']['error']['text'] = error_msg
            msg['failure']['error']['nodeId'] = nodeId
            msg['failure']['error']['timestamp'] = datetime.datetime.now().replace(microsecond=0).isoformat()

            # Drop communication with this device and check if we are done
            self.sessions[session]["nodeDone"][nodeId] = True
            if (self._all_nodes_done(session)):
                msg['failure']['done'] = 'true'
                # The session is complete, delete it
                del self.sessions[session]
            msg.send()
        else:
            msg = self.xmpp.Message()
            msg['from'] = self.sessions[session]['to']
            msg['to'] = self.sessions[session]['from']
            msg['fields']['seqnr'] = self.sessions[session]['seqnr']

            if timestamp_block is not None and len(timestamp_block) > 0:
                node = msg['fields'].add_node(nodeId)
                ts = node.add_timestamp(timestamp_block["timestamp"])

                for f in timestamp_block["fields"]:
                    data = ts.add_data( typename=f['type'],
                                        name=f['name'],
                                        value=f['value'],
                                        unit=f['unit'],
                                        dataType=f['dataType'],
                                        flags=f['flags'])

            if result == "done":
                self.sessions[session]["commTimers"][nodeId].cancel()
                self.sessions[session]["nodeDone"][nodeId] = True
                if (self._all_nodes_done(session)):
                    # The session is complete, delete it
                    del self.sessions[session]
                    msg['fields']['done'] = 'true'
            else:
                # Restart comm timer
                self.sessions[session]["commTimers"][nodeId].reset()

            msg.send()

    def _handle_event_cancel(self, iq):
        """ Received Iq with cancel - this is a cancel request.
        Delete the session and confirm. """

        seqnr = iq['cancel']['seqnr']
        # Find the session
        for s in self.sessions:
            if self.sessions[s]['from'] == iq['from'] and self.sessions[s]['to'] == iq['to'] and self.sessions[s]['seqnr'] == seqnr:
                # found it. Cancel all timers
                for n in self.sessions[s]["commTimers"]:
                    self.sessions[s]["commTimers"][n].cancel()

                # Confirm
                iq = iq.reply()
                iq['type'] = 'result'
                iq['cancelled']['seqnr'] = seqnr
                iq.send()

                # Delete session
                del self.sessions[s]
                return

        # Could not find session, send reject
        iq = iq.reply()
        iq['type'] = 'error'
        iq['rejected']['seqnr'] = seqnr
        iq['rejected']['error'] = "Cancel request received, no matching request is active."
        iq.send()

        # =================================================================
    # Client side (data retriever) API

    def request_data(self, from_jid, to_jid, callback, nodeIds=None, fields=None, flags=None):
        """
        Called on the client side to initiate a data readout.
        Composes a message with the request and sends it to the device(s).
        Does not block, the callback will be called when data is available.

        Arguments:
            from_jid        -- The jid of the requester
            to_jid          -- The jid of the device(s)
            callback        -- The callback function to call when data is available.

                            The callback function must support the following arguments:

                from_jid    -- The jid of the responding device(s)
                result      -- The current result status of the readout. Valid values are:
                               "accepted"  - Readout request accepted
                               "queued"    - Readout request accepted and queued
                               "rejected"  - Readout request rejected
                               "failure"   - Readout failed.
                               "cancelled" - Confirmation of request cancellation.
                               "started"   - Previously queued request is now started
                               "fields"    - Contains readout data.
                               "done"      - Indicates that the readout is complete.

                nodeId      -- [optional] Mandatory when result == "fields" or "failure".
                               The node Id of the responding device. One callback will only
                               contain data from one device.
                timestamp   -- [optional] Mandatory when result == "fields".
                               The timestamp of data in this callback. One callback will only
                               contain data from one timestamp.
                fields      -- [optional] Mandatory when result == "fields".
                               List of field dictionaries representing the readout data.
                               Dictionary format:
                  {
                    typename:  The field type (numeric, boolean, dateTime, timeSpan, string, enum)
                    name:      The field name
                    value:     The field value
                    unit:      The unit of the field. Only applies to type numeric.
                    dataType:  The datatype of the field. Only applies to type enum.
                    flags:     [optional] data classifier flags for the field, e.g. momentary.
                               Formatted as a dictionary like { "flag name": "flag value" ... }
                  }

                error_msg   -- [optional] Mandatory when result == "rejected" or "failure".
                               Details about why the request is rejected or failed.
                               "rejected" means that the request is stopped, but note that the
                               request will continue even after a "failure". "failure" only means
                               that communication was stopped to that specific device, other
                               device(s) (if any) will continue their readout.

            nodeIds      -- [optional] Limits the request to the node Ids in this list.
            fields       -- [optional] Limits the request to the field names in this list.
            flags        -- [optional] Limits the request according to the flags, or sets
                            readout conditions such as timing.

        Return value:
            session      -- Session identifier. Client can use this as a reference to cancel
                            the request.
        """
        iq = self.xmpp.Iq()
        iq['from'] = from_jid
        iq['to'] = to_jid
        iq['type'] = "get"
        seqnr = self._get_new_seqnr()
        iq['id'] = seqnr
        iq['req']['seqnr'] = seqnr
        if nodeIds is not None:
            for nodeId in nodeIds:
                iq['req'].add_node(nodeId)
        if fields is not None:
            for field in fields:
                iq['req'].add_field(field)

        iq['req']._set_flags(flags)

        self.sessions[seqnr] = {"from": iq['from'], "to": iq['to'], "seqnr": seqnr, "callback": callback}
        iq.send()

        return seqnr

    def cancel_request(self, session):
        """
        Called on the client side to cancel a request for data readout.
        Composes a message with the cancellation and sends it to the device(s).
        Does not block, the callback will be called when cancellation is
        confirmed.

        Arguments:
            session        -- The session id of the request to cancel
        """
        seqnr = session
        iq = self.xmpp.Iq()
        iq['from'] = self.sessions[seqnr]['from']
        iq['to'] = self.sessions[seqnr]['to']
        iq['type'] = "get"
        iq['id'] = seqnr
        iq['cancel']['seqnr'] = seqnr
        iq.send()

    def _get_new_seqnr(self):
        """ Returns a unique sequence number (unique across threads) """
        self.seqnr_lock.acquire()
        self.last_seqnr += 1
        self.seqnr_lock.release()
        return str(self.last_seqnr)

    def _handle_event_accepted(self, iq):
        """ Received Iq with accepted - request was accepted """
        seqnr = iq['accepted']['seqnr']
        result = "accepted"
        if iq['accepted']['queued'] == 'true':
            result = "queued"

        callback = self.sessions[seqnr]["callback"]
        callback(from_jid=iq['from'], result=result)

    def _handle_event_rejected(self, iq):
        """ Received Iq with rejected - this is a reject.
        Delete the session. """
        seqnr = iq['rejected']['seqnr']
        callback = self.sessions[seqnr]["callback"]
        callback(from_jid=iq['from'], result="rejected", error_msg=iq['rejected']['error'])
        # Session terminated
        del self.sessions[seqnr]

    def _handle_event_cancelled(self, iq):
        """
        Received Iq with cancelled - this is a cancel confirm.
        Delete the session.
        """
        seqnr = iq['cancelled']['seqnr']
        callback = self.sessions[seqnr]["callback"]
        callback(from_jid=iq['from'], result="cancelled")
        # Session cancelled
        del self.sessions[seqnr]

    def _handle_event_fields(self, msg):
        """
        Received Msg with fields - this is a data response to a request.
        If this is the last data block, issue a "done" callback.
        """
        seqnr = msg['fields']['seqnr']
        callback = self.sessions[seqnr]["callback"]
        for node in msg['fields']['nodes']:
            for ts in node['timestamps']:
                fields = []
                for d in ts['datas']:
                    field_block = {}
                    field_block["name"] = d['name']
                    field_block["typename"] = d._get_typename()
                    field_block["value"] = d['value']
                    if not d['unit'] == "": field_block["unit"] = d['unit']
                    if not d['dataType'] == "": field_block["dataType"] = d['dataType']
                    flags = d._get_flags()
                    if not len(flags) == 0:
                        field_block["flags"] = flags
                    fields.append(field_block)

                callback(from_jid=msg['from'], result="fields", nodeId=node['nodeId'], timestamp=ts['value'], fields=fields)

        if msg['fields']['done'] == "true":
            callback(from_jid=msg['from'], result="done")
            # Session done
            del self.sessions[seqnr]

    def _handle_event_failure(self, msg):
        """
        Received Msg with failure - our request failed
        Delete the session.
        """
        seqnr = msg['failure']['seqnr']
        callback = self.sessions[seqnr]["callback"]
        callback(from_jid=msg['from'], result="failure", nodeId=msg['failure']['error']['nodeId'], timestamp=msg['failure']['error']['timestamp'], error_msg=msg['failure']['error']['text'])

        # Session failed
        del self.sessions[seqnr]

    def _handle_event_started(self, msg):
        """
        Received Msg with started - our request was queued and is now started.
        """
        seqnr = msg['started']['seqnr']
        callback = self.sessions[seqnr]["callback"]
        callback(from_jid=msg['from'], result="started")
